from pprint import pprint

from ..pallet import Pallet
from .util import ELiDEAppTest, idle_until, window_with_widget


class TestSpriteBuilder(ELiDEAppTest):

	def test_build_pawn(self):
		app = self.app
		win = window_with_widget(app.build())
		app.manager.current = 'pawncfg'
		idle_until(lambda: 'dialog' in app.pawncfg.ids, 100,
					"Never made dialog for pawncfg")
		pawn_cfg_dialog = app.pawncfg.ids.dialog
		idle_until(lambda: 'builder' in pawn_cfg_dialog.ids, 100,
					"Never made pawn builder")
		builder = pawn_cfg_dialog.ids.builder
		idle_until(lambda: builder.labels, 100, "Never got any builder labels")
		idle_until(lambda: builder.pallets, 100,
					"Never got any builder pallets")
		idle_until(lambda: len(builder.labels) == len(builder.pallets), 100,
					"Never updated pawn builder")
		palbox = builder._palbox
		for child in palbox.children:
			if not isinstance(child, Pallet):
				continue
			idle_until(lambda: child.swatches, 100,
						"Never got swatches for " + child.filename)
			if 'draconian_m' in child.swatches:
				child.swatches['draconian_m'].state = 'down'
				idle_until(
					lambda: child.swatches['draconian_m'] in child.selection,
					100, "Selection never updated")
			if 'robe_red' in child.swatches:
				child.swatches['robe_red'].state = 'down'
				idle_until(
					lambda: child.swatches['robe_red'] in child.selection, 100,
					"Selection never updated")
		idle_until(lambda: pawn_cfg_dialog.ids.selector.imgpaths, 100,
					"Never got imgpaths")
		pawn_cfg_dialog.pressed()
		idle_until(lambda: pawn_cfg_dialog.imgpaths, 100,
					"Never propagated imgpaths")
		assert pawn_cfg_dialog.imgpaths == [
			'atlas://base.atlas/draconian_m', 'atlas://body.atlas/robe_red'
		]



# This file is part of ELiDE, frontend to LiSE, a framework for life simulation games.
# Copyright (c) Zachary Spector, public@zacharyspector.com
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, version 3.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""That which displays a one-way connection between two places.

An arrow connects two spots, the origin and the destination, and it
points from the origin to the destination, regardless of where on the
screen they are at the moment.

"""
from math import cos, sin, atan, pi
import numpy as np
from kivy.uix.widget import Widget
from kivy.uix.floatlayout import FloatLayout
from kivy.graphics.fbo import Fbo
from kivy.graphics import (Translate, Rectangle, Quad, Color)
from kivy.properties import (ReferenceListProperty, AliasProperty,
								ObjectProperty, NumericProperty, ListProperty,
								BooleanProperty, StringProperty)
from kivy.clock import Clock

from ..dummy import Dummy
from ..util import trigger

try:
	from kivy.garden.collider import Collide2DPoly
except (KeyError, ImportError):
	from ..collide import Collide2DPoly
from ..util import get_thin_rect_vertices, fortyfive

cos45 = cos(fortyfive)
sin45 = sin(fortyfive)


def up_and_down(orig, dest, taillen):
	if orig.center_y == dest.center_y:
		raise ValueError("Can't draw an arrow at a point")
	flipped = orig.center_y > dest.center_y
	if flipped:
		orig, dest = dest, orig
	x = int(orig.center_x)
	dy = int(dest.y)
	oy = int(orig.top)
	if flipped:
		oy, dy = dy, oy
	off1 = cos45 * taillen
	off2 = sin45 * taillen
	x0 = x
	y0 = oy
	endx = x
	endy = dy
	x1 = endx - off1
	x2 = endx + off1
	y1 = y2 = endy - off2 if oy < dy else endy + off2
	return ([x0, y0, endx, endy], [x1, y1, endx, endy, x2, y2])


def left_and_right(orig, dest, taillen):
	if orig.center_x == dest.center_x:
		raise ValueError("Can't draw an arrow at a point")
	flipped = orig.center_x > dest.center_x
	if flipped:
		orig, dest = dest, orig
	y = int(orig.center_y)
	dx = int(dest.x)
	ox = int(orig.right)
	if flipped:
		ox, dx = dx, ox
	off1 = cos45 * taillen
	off2 = sin45 * taillen
	x0 = ox
	y0 = y
	endx = dx
	endy = y
	y1 = endy - off1
	y2 = endy + off1
	x1 = x2 = endx - off2 if ox < dx else endx + off2
	return ([x0, y0, endx, endy], [x1, y1, endx, endy, x2, y2])


def _get_points_first_part(orig, dest, taillen):
	ox, oy = orig.center
	ow, oh = orig.size
	dx, dy = dest.center
	dw, dh = dest.size
	# Flip the endpoints around so that the arrow faces up and right.
	# We'll flip it back at the end as needed.
	xco = 1 if ox < dx else -1
	yco = 1 if oy < dy else -1
	ox *= xco
	dx *= xco
	oy *= yco
	dy *= yco
	# Nudge my endpoints so that I start and end at the edge of
	# a Spot.
	if dy - oy > dx - ox:
		topy = dy - dh / 2
		boty = oy + oh / 2
		leftx = ox
		rightx = dx
	else:
		leftx = ox + ow / 2
		rightx = dx - dw / 2
		topy = dy
		boty = oy
	# Degenerate cases.
	# Also, these need to be handled specially to avoid
	# division by zero later on.
	if rightx == leftx:
		return up_and_down(orig, dest, taillen)
	if topy == boty:
		return left_and_right(orig, dest, taillen)
	return ow, oh, dw, dh, xco, leftx, rightx, yco, topy, boty


def get_points_multi(args):
	"""Return a dictionary mapping (orig, dest) to pairs of point lists for arrows

    Takes an iterable of (orig, dest, taillen) where orig and dest are Spot instances

    taillen is an integer specifying how long the arrowhead should be.

    """
	ret = {}
	keys = []
	topys = []
	botys = []
	leftxs = []
	rightxs = []
	taillens = []
	xcos = []
	ycos = []
	for (orig, dest, taillen) in args:
		try:
			p1 = _get_points_first_part(orig, dest, taillen)
		except ValueError:
			p1 = 2, 2, 2, 2, 1, 0, 1, 1, 0, 1
		if len(p1) == 2:
			ret[orig, dest] = p1
			continue
		ow, oh, dw, dh, xco, leftx, rightx, yco, topy, boty = p1
		keys.append((orig, dest))
		leftxs.append(leftx)
		rightxs.append(rightx)
		topys.append(topy)
		botys.append(boty)
		taillens.append(taillen)
		xcos.append(xco)
		ycos.append(yco)
	topys = np.array(topys)
	botys = np.array(botys)
	rightxs = np.array(rightxs)
	leftxs = np.array(leftxs)
	rises = np.subtract(topys, botys)
	runs = np.subtract(rightxs, leftxs)
	slopes = np.divide(rises, runs)
	unslopes = np.divide(runs, rises)
	start_thetas = np.arctan(slopes)
	end_thetas = np.arctan(unslopes)
	top_thetas = np.subtract(start_thetas, fortyfive)
	bot_thetas = np.subtract(pi - fortyfive, end_thetas)
	topsins = np.sin(top_thetas)
	topcoss = np.cos(top_thetas)
	botsins = np.sin(bot_thetas)
	botcoss = np.cos(bot_thetas)
	taillens = np.array(taillens)
	xoff1s = np.multiply(topcoss, taillens)
	yoff1s = np.multiply(topsins, taillens)
	xoff2s = np.multiply(botcoss, taillens)
	yoff2s = np.multiply(botsins, taillens)
	xcos = np.array(xcos)
	ycos = np.array(ycos)
	x1s = np.multiply(np.subtract(rightxs, xoff1s), xcos)
	x2s = np.multiply(np.subtract(rightxs, xoff2s), xcos)
	y1s = np.multiply(np.subtract(topys, yoff1s), ycos)
	y2s = np.multiply(np.subtract(topys, yoff2s), ycos)
	startxs = np.multiply(leftxs, xcos)
	startys = np.multiply(botys, ycos)
	endxs = np.multiply(rightxs, xcos)
	endys = np.multiply(topys, ycos)
	for key, startx, starty, endx, endy, x1, y1, endx, endy, x2, y2 in zip(
		keys, startxs, startys, endxs, endys, x1s, y1s, endxs, endys, x2s,
		y2s):
		ret[key] = ([startx, starty, endx, endy], [x1, y1, endx, endy, x2, y2])
	return ret


def get_points(orig, dest, taillen):
	"""Return a pair of lists of points for use making an arrow.

    The first list is the beginning and end point of the trunk of the arrow.

    The second list is the arrowhead.

    """
	p1 = _get_points_first_part(orig, dest, taillen)
	if len(p1) == 2:
		return p1
	ow, oh, dw, dh, xco, leftx, rightx, yco, topy, boty = p1
	rise = topy - boty
	run = rightx - leftx
	slope = rise / run
	unslope = run / rise
	start_theta = atan(slope)
	end_theta = atan(unslope)

	# make the little wedge at the end so you can tell which way the
	# arrow's pointing, and flip it all back around to the way it was
	top_theta = start_theta - fortyfive
	bot_theta = pi - fortyfive - end_theta
	xoff1 = cos(top_theta) * taillen
	yoff1 = sin(top_theta) * taillen
	xoff2 = cos(bot_theta) * taillen
	yoff2 = sin(bot_theta) * taillen
	x1 = (rightx - xoff1) * xco
	x2 = (rightx - xoff2) * xco
	y1 = (topy - yoff1) * yco
	y2 = (topy - yoff2) * yco
	startx = leftx * xco
	starty = boty * yco
	endx = rightx * xco
	endy = topy * yco
	return ([startx, starty, endx, endy], [x1, y1, endx, endy, x2, y2])


eight0s = tuple([0] * 8)


class GraphArrowWidget(Widget):
	"""A widget that points from one :class:`~LiSE.gui.graph.Spot` to
    another.

    :class:`Arrow`s are the graphical representations of
    :class:`~LiSE.model.Portal`s. They point from the :class:`Spot`
    representing the :class:`Portal`'s origin, to the one representing
    its destination.

    """
	board = ObjectProperty()
	name = StringProperty()
	margin = NumericProperty(10)
	"""When deciding whether a touch collides with me, how far away can
    the touch get before I should consider it a miss?"""
	w = NumericProperty(2)
	"""The width of the inner, brighter portion of the :class:`Arrow`. The
    whole :class:`Arrow` will end up thicker."""
	pawns_here = ListProperty()
	trunk_points = ListProperty()
	head_points = ListProperty()
	points = ReferenceListProperty(trunk_points, head_points)
	trunk_quad_vertices_bg = ListProperty(eight0s)
	trunk_quad_vertices_fg = ListProperty(eight0s)
	left_head_quad_vertices_bg = ListProperty(eight0s)
	right_head_quad_vertices_bg = ListProperty(eight0s)
	left_head_quad_vertices_fg = ListProperty(eight0s)
	right_head_quad_vertices_fg = ListProperty(eight0s)
	slope = NumericProperty(0.0, allownone=True)
	origin = ObjectProperty()
	destination = ObjectProperty()
	repointed = BooleanProperty(True)
	bg_scale_unselected = NumericProperty(4)
	bg_scale_selected = NumericProperty(5)
	selected = BooleanProperty(False)
	bg_color_unselected = ListProperty([0.5, 0.5, 0.5, 0.5])
	bg_color_selected = ListProperty([0.0, 1.0, 1.0, 1.0])
	fg_color_unselected = ListProperty([1.0, 1.0, 1.0, 1.0])
	fg_color_selected = ListProperty([1.0, 1.0, 1.0, 1.0])
	bg_color_unselected_head = ListProperty()
	bg_color_selected_head = ListProperty()
	fg_color_unselected_head = ListProperty()
	fg_color_selected_head = ListProperty()
	arrowhead_size = NumericProperty(10)
	collide_radius = NumericProperty(3)
	collider = ObjectProperty()
	portal = ObjectProperty()

	def on_portal(self, *args):
		"""Set my ``name`` and instantiate my ``mirrormap`` as soon as I have
        the properties I need to do so.

        """
		if not (self.board and self.origin and self.destination
				and self.origin.name in self.board.character.portal
				and self.destination.name in self.board.character.portal):
			Clock.schedule_once(self.on_portal, 0)
			return
		self.name = '{}->{}'.format(self.portal['origin'],
									self.portal['destination'])

	def collide_point(self, x, y):
		"""Delegate to my ``collider``, or return ``False`` if I don't have
        one.

        """
		if not self.collider:
			return False
		return (x, y) in self.collider

	def __init__(self, **kwargs):
		"""Create trigger for my _repoint method. Delegate to parent for
        everything else.

        """
		self._trigger_repoint = Clock.create_trigger(self._repoint, timeout=-1)
		super().__init__(**kwargs)

	def on_origin(self, *args):
		"""Make sure to redraw whenever the origin moves."""
		if self.origin is None:
			Clock.schedule_once(self.on_origin, 0)
			return
		if hasattr(self, '_origin'):
			if hasattr(self._origin, '_bound_pos_repoint'):
				self._origin.unbind_uid('pos', self._origin._bound_pos_repoint)
				del self._origin._bound_pos_repoint
			if hasattr(self._origin, '_bound_size_repoint'):
				self._origin_unbind_uid('size',
										self._origin._bound_size_repoint)
				del self._origin._bound_size
		origin = self._origin = self.origin
		origin._bound_pos_repoint = origin.fbind('pos', self._trigger_repoint)
		origin._bound_size_repoint = origin.fbind('size',
													self._trigger_repoint)

	def on_destination(self, *args):
		"""Make sure to redraw whenever the destination moves."""
		if self.destination is None:
			Clock.schedule_once(self.on_destination, 0)
			return
		if hasattr(self, '_destination'):
			if hasattr(self._destination, '_bound_pos_repoint'):
				self._destination.unbind_uid(
					'pos', self._destination._bound_pos_repoint)
				del self._destination._bound_pos_repoint
			if hasattr(self._destination, '_bound_size_repoint'):
				self.destination.unbind_uid(
					'size', self._destination._bound_size_repoint)
				del self._destination._bound_size_repoint
		destination = self._destination = self.destination
		destination._bound_pos_repoint = destination.fbind(
			'pos', self._trigger_repoint)
		destination._bound_size_repoint = destination.fbind(
			'size', self._trigger_repoint)

	def add_widget(self, wid, index=0, canvas=None):
		"""Put the :class:`Pawn` at a point along my length proportionate to
        how close it is to finishing its travel through me.

        Only :class:`Pawn` should ever be added as a child of :class:`Arrow`.

        """
		super().add_widget(wid, index, canvas)
		if not hasattr(wid, 'group'):
			return
		wid._no_use_canvas = True
		mycanvas = (self.canvas.before if canvas == 'before' else
					self.canvas.after if canvas == 'after' else self.canvas)
		mycanvas.remove(wid.canvas)
		pawncanvas = (self.board.spotlayout.canvas.before if canvas == 'before'
						else self.board.spotlayout.canvas.after
						if canvas == 'after' else self.board.spotlayout.canvas)
		for child in self.children:
			if hasattr(child, 'group') and child.group in pawncanvas.children:
				pawncanvas.remove(child.group)
			pawncanvas.add(child.group)
		self.pospawn(wid)

	def remove_widget(self, wid):
		"""Remove the special :class:`InstructionGroup` I was using to draw
        this :class:`Pawn`.

        """
		super().remove_widget(wid)
		wid._no_use_canvas = False

	def on_points(self, *args):
		"""Reposition my children when I have new points."""
		for pawn in self.children:
			self.pospawn(pawn)

	def pos_along(self, pct):
		"""Return coordinates for where a Pawn should be if it has travelled
        along ``pct`` of my length (between 0 and 1).

        Might get complex when I switch over to using beziers for
        arrows, but for now this is quite simple, using distance along
        a line segment.

        """
		if pct < 0 or pct > 1:
			raise ValueError("Invalid portion")
		(ox, oy) = self.origin.center
		(dx, dy) = self.destination.center
		xdist = (dx - ox) * pct
		ydist = (dy - oy) * pct
		return (ox + xdist, oy + ydist)

	def pospawn(self, pawn):
		"""Position a :class:`Pawn` that is my child so as to reflect how far
        its :class:`Thing` has gone along my :class:`Portal`.

        """
		if self._turn < pawn.thing['arrival_time']:
			# It's weird that the pawn is getting placed in me, but
			# I'll do my best..
			pawn.pos = self.pos_along(0)
			return
		elif (pawn.thing['next_arrival_time']
				and self._turn >= pawn.thing['next_arrival_time']):
			pawn.pos = self.pos_along(1)
			return
		try:
			pawn.pos = self.pos_along(
				(self._turn - pawn.thing['arrival_time']) /
				(pawn.thing['next_arrival_time'] - pawn.thing['arrival_time']))
		except (TypeError, ZeroDivisionError):
			pawn.pos = self.pos_along(0)

	def _get_points(self):
		"""Return the coordinates of the points that describe my shape."""
		return get_points(self.origin, self.destination, self.arrowhead_size)

	def _get_slope(self):
		"""Return a float of the increase in y divided by the increase in x,
        both from left to right."""
		orig = self.origin
		dest = self.destination
		ox = orig.x
		oy = orig.y
		dx = dest.x
		dy = dest.y
		if oy == dy:
			return 0
		elif ox == dx:
			return None
		else:
			rise = dy - oy
			run = dx - ox
			return rise / run

	def _get_b(self):
		"""Return my Y-intercept.

        I probably don't really hit the left edge of the window, but
        this is where I would, if I were long enough.

        """
		orig = self.origin
		dest = self.destination
		(ox, oy) = orig.pos
		(dx, dy) = dest.pos
		denominator = dx - ox
		x_numerator = (dy - oy) * ox
		y_numerator = denominator * oy
		return ((y_numerator - x_numerator), denominator)

	def _repoint(self, *args):
		"""Recalculate points, y-intercept, and slope"""
		if None in (self.origin, self.destination):
			Clock.schedule_once(self._repoint, 0)
			return
		try:
			(self.trunk_points, self.head_points) = self._get_points()
		except ValueError:
			self.trunk_points = self.head_points = []
			return
		(ox, oy, dx, dy) = self.trunk_points
		r = self.w / 2
		bgr = r * self.bg_scale_selected if self.selected \
               else self.bg_scale_unselected
		self.trunk_quad_vertices_bg = get_thin_rect_vertices(
			ox, oy, dx, dy, bgr)
		self.collider = Collide2DPoly(self.trunk_quad_vertices_bg)
		self.trunk_quad_vertices_fg = get_thin_rect_vertices(ox, oy, dx, dy, r)
		(x1, y1, endx, endy, x2, y2) = self.head_points
		self.left_head_quad_vertices_bg = get_thin_rect_vertices(
			x1, y1, endx, endy, bgr)
		self.right_head_quad_vertices_bg = get_thin_rect_vertices(
			x2, y2, endx, endy, bgr)
		self.left_head_quad_vertices_fg = get_thin_rect_vertices(
			x1, y1, endx, endy, r)
		self.right_head_quad_vertices_fg = get_thin_rect_vertices(
			x2, y2, endx, endy, r)
		self.slope = self._get_slope()
		self.repointed = True

	@trigger
	def _pull_bg_color0(self, *args):
		self._color0.rgba = self.bg_color_selected if self.selected else self.bg_color_unselected

	@trigger
	def _pull_points_quad0(self, *args):
		self._quad0.points = self.trunk_quad_vertices_bg

	@trigger
	def _pull_head_bg_color1(self, *args):
		if self.selected:
			self._color1.rgba = self.bg_color_selected_head or self.bg_color_selected
		else:
			self._color1.rgba = self.bg_color_unselected_head or self.bg_color_unselected

	@trigger
	def _pull_bg_left_head_points_quad1_0(self, *args):
		self._quad1_0.points = self.left_head_quad_vertices_bg

	@trigger
	def _pull_bg_right_head_points_quad1_1(self, *args):
		self._quad1_1.points = self.right_head_quad_vertices_bg

	@trigger
	def _pull_color2(self, *args):
		if self.selected:
			self._color2.rgba = self.fg_color_selected
		else:
			self._color2.rgba = self.fg_color_unselected

	@trigger
	def _pull_color3(self, *args):
		if self.selected:
			self._color3.rgba = self.fg_color_selected_head or self.fg_color_selected
		else:
			self._color3.rgba = self.fg_color_unselected_head or self.fg_color_unselected

	@trigger
	def _pull_quad2_points(self, *args):
		self._quad2.points = self.trunk_quad_vertices_fg

	@trigger
	def _pull_points_quad3_0(self, *args):
		self._quad3_0.points = self.left_head_quad_vertices_fg

	@trigger
	def _pull_points_quad3_1(self, *args):
		self._quad3_1.points = self.right_head_quad_vertices_fg

	def on_parent(self, *args):
		if not self.canvas:
			Clock.schedule_once(self.on_parent, 0)
			return
		if not self.trunk_points or not self.head_points:
			self._repoint()
		if hasattr(self, '_color0'):
			return
		with self.canvas:
			self._color0 = Color(rgba=self.bg_color_selected if self.
									selected else self.bg_color_unselected)
			for att in (
				'bg_color_selected',
				'bg_color_unselected',
				'selected',
			):
				self.fbind(att, self._pull_bg_color0)
			self._quad0 = Quad(points=self.trunk_quad_vertices_bg)
			self.fbind('trunk_quad_vertices_bg', self._pull_points_quad0)
			self._color1 = Color(rgba=(
				self.bg_color_selected_head or self.bg_color_selected
			) if self.selected else (
				self.bg_color_unselected_head or self.bg_color_unselected))
			for att in ('bg_color_selected_head', 'bg_color_selected',
						'bg_color_unselected_head', 'bg_color_unselected',
						'selected'):
				self.fbind(att, self._pull_head_bg_color1)
			self._quad1_0 = Quad(points=self.left_head_quad_vertices_bg)
			self.fbind('left_head_quad_vertices_bg',
						self._pull_bg_left_head_points_quad1_0)
			self._quad1_1 = Quad(points=self.right_head_quad_vertices_bg)
			self.fbind('right_head_quad_vertices_bg',
						self._pull_bg_right_head_points_quad1_1)
			self._color2 = Color(rgba=self.fg_color_selected if self.
									selected else self.fg_color_unselected)
			for att in ('fg_color_selected', 'fg_color_unselected',
						'selected'):
				self.fbind(att, self._pull_color2)
			self._quad2 = Quad(points=self.trunk_quad_vertices_fg)
			self.fbind('trunk_quad_vertices_fg', self._pull_quad2_points)
			self._color3 = Color(rgba=(
				self.fg_color_selected_head or self.fg_color_selected
			) if self.selected else (
				self.fg_color_unselected_head or self.fg_color_unselected))
			for att in ('selected', 'fg_color_selected_head',
						'fg_color_selected', 'fg_color_unselected_head',
						'fg_color_unselected'):
				self.fbind(att, self._pull_color3)
			self._quad3_0 = Quad(points=self.left_head_quad_vertices_fg)
			self.fbind('left_head_quad_vertices_fg', self._pull_points_quad3_0)
			self._quad3_1 = Quad(points=self.right_head_quad_vertices_fg)
			self.fbind('right_head_quad_vertices_fg',
						self._pull_points_quad3_1)


class GraphArrow(GraphArrowWidget):
	"""A :class:`GraphArrowWidget` that represents a :class:`Portal` object.

    This subclass is much more often used than :class:`GraphArrowWidget`,
    which is only seen on its own when the user is in the process of
    creating a new :class:`Portal`.

    """
	origspot = ObjectProperty()
	destspot = ObjectProperty()
	portal = ObjectProperty()
	"""The portal that I represent."""
	reciprocal = AliasProperty(lambda self: self._get_reciprocal()
								if None not in
								(self.board, self.portal) else None,
								lambda self, v: None,
								bind=('portal', ))
	grabbed = BooleanProperty(False)
	_turn = NumericProperty()

	def _get_reciprocal(self):
		"""Return the :class:`Arrow` that connects my origin and destination
        in the opposite direction, if it exists.

        """
		orign = self.portal['origin']
		destn = self.portal['destination']
		if (destn in self.board.arrow and orign in self.board.arrow[destn]):
			return self.board.arrow[destn][orign]
		else:
			return None

	def on_portal(self, *args):
		orig = self.portal['origin']
		dest = self.portal['destination']
		spot = self.board.spot
		self.origin = spot[orig] if orig in spot else Dummy()
		self.destination = spot[dest] if dest in spot else Dummy()


class ArrowLayout(FloatLayout):

	def __init__(self, **kwargs):
		self._trigger_redraw = Clock.create_trigger(self.redraw)
		self.bind(children=self._trigger_redraw)
		super().__init__(**kwargs)

	def on_parent(self, *args):
		if not self.canvas:
			Clock.schedule_once(self.on_parent, 0)
			return
		with self.canvas:
			self._fbo = Fbo(size=self.size)
			self._translate = Translate(x=self.x, y=self.y)
			self._rectangle = Rectangle(size=self.size,
										texture=self._fbo.texture)

	def redraw(self, *args):
		if not hasattr(self, '_rectangle'):
			self._trigger_redraw()
			return
		fbo = self._fbo
		fbo.bind()
		fbo.clear()
		fbo.clear_buffer()
		fbo.release()
		trigger_redraw = self._trigger_redraw
		redraw_bound = '_redraw_bound_' + str(id(self))
		for child in self.children:
			if hasattr(child, redraw_bound):
				child.unbind_uid('selected', getattr(child, redraw_bound))
			fbo.add(child.canvas)
			setattr(child, redraw_bound, child.fbind('selected',
														trigger_redraw))
			if hasattr(child.origspot, redraw_bound):
				child.origspot.unbind_uid(
					'pos', getattr(child.origspot, redraw_bound))
			setattr(child.origspot, redraw_bound,
					child.origspot.fbind('pos', trigger_redraw))
			if hasattr(child.destspot, redraw_bound):
				child.destspot.unbind_uid(
					'pos', getattr(child.destspot, redraw_bound))
			setattr(child.destspot, redraw_bound,
					child.destspot.fbind('pos', trigger_redraw))

	def on_pos(self, *args):
		if not hasattr(self, '_translate'):
			return
		self._translate.x, self._translate.y = self.pos

	def on_size(self, *args):
		if not hasattr(self, '_rectangle') or not hasattr(self, '_fbo'):
			return
		self._rectangle.size = self._fbo.size = self.size
		self.redraw()

	def add_widget(self, widget, index=0, canvas=None):
		ret = super().add_widget(widget, index, canvas)
		if canvas is None:
			self.canvas.children.remove(widget.canvas)
		self._trigger_redraw()
		return ret

	def remove_widget(self, widget):
		super().remove_widget(widget)
		self._trigger_redraw()



from LiSE import Engine
from LiSE.examples import kobold
from .util import ELiDEAppTest, window_with_widget, idle_until


class TestPythonEditor(ELiDEAppTest):

	def setUp(self):
		super().setUp()
		with Engine(self.prefix) as eng:
			kobold.inittest(eng)

	def _get_actions_box(self):
		app = self.app
		idle_until(
			lambda: hasattr(app, 'mainscreen') and app.mainscreen.mainview and
			app.mainscreen.statpanel and hasattr(app.mainscreen, 'gridview'))
		app.funcs.toggle()
		idle_until(lambda: 'actions' in app.funcs.ids, 100,
					'Never got actions box')
		actions_box = app.funcs.ids.actions
		idle_until(lambda: actions_box.editor, 100, 'Never got FuncEditor')
		idle_until(lambda: actions_box.storelist, 100, 'Never got StoreList')
		idle_until(lambda: actions_box.storelist.data, 100,
					'Never got StoreList data')
		return actions_box

	def test_show_code(self):
		app = self.app
		win = window_with_widget(app.build())
		actions_box = self._get_actions_box()
		last = actions_box.storelist.data[-1]['name']
		actions_box.storelist.selection_name = last
		idle_until(lambda: 'funname' in actions_box.editor.ids, 100,
					'Never got function input widget')
		idle_until(lambda: actions_box.editor.ids.funname.hint_text, 100,
					"Never got function name")
		idle_until(lambda: 'code' in actions_box.editor.ids, 100,
					"Never got code editor widget")
		idle_until(lambda: actions_box.editor.ids.code.text, 100,
					"Never got source code")

	def test_create_action(self):
		app = self.app
		win = window_with_widget(app.build())
		actions_box = self._get_actions_box()
		actions_box.editor.ids.funname.text = 'new_func'
		actions_box.editor.ids.code.text = 'return "Hello, world!"'
		app.stop()
		with Engine(self.prefix) as eng:
			assert hasattr(eng.action, 'new_func')



import sys
from tempfile import mkdtemp
import shutil
from functools import partial

from blinker import Signal
from kivy.input.motionevent import MotionEvent
from kivy.base import EventLoop
from kivy.tests.common import GraphicUnitTest
from kivy.config import ConfigParser

from LiSE.proxy import RedundantProcessError
from ELiDE.app import ELiDEApp


def all_spots_placed(board, char=None):
	if char is None:
		char = board.character
	for place in char.place:
		if place not in board.spot:
			return False
	return True


def all_pawns_placed(board, char=None):
	if char is None:
		char = board.character
	for thing in char.thing:
		if thing not in board.pawn:
			return False
	return True


def all_arrows_placed(board, char=None):
	if char is None:
		char = board.character
	for orig, dests in char.portal.items():
		if orig not in board.arrow:
			return False
		arrows = board.arrow[orig]
		for dest in dests:
			if dest not in arrows:
				return False
	return True


def board_is_arranged(board, char=None):
	if char is None:
		char = board.character
	return all_spots_placed(board, char) and all_pawns_placed(
		board, char) and all_arrows_placed(board, char)


def idle_until(condition=None, timeout=100, message="Timed out"):
	"""Advance frames until ``condition()`` is true

    With integer ``timeout``, give up after that many frames,
    raising ``TimeoutError``. You can customize its ``message``.

    """
	if not (timeout or condition):
		raise ValueError("Need timeout or condition")
	if condition is None:
		return partial(idle_until, timeout=timeout, message=message)
	if timeout is None:
		while not condition():
			EventLoop.idle()
		return
	for _ in range(timeout):
		if condition():
			return
		EventLoop.idle()
	raise TimeoutError(message)


def window_with_widget(wid):
	EventLoop.ensure_window()
	win = EventLoop.window
	win.add_widget(wid)
	return win


class MockTouch(MotionEvent):

	def depack(self, args):
		self.is_touch = True
		self.sx = args['sx']
		self.sy = args['sy']
		super().depack(args)


class ListenableDict(dict, Signal):

	def __init__(self):
		Signal.__init__(self)


class MockTime(Signal):
	pass


class MockEngine(Signal):
	eternal = ListenableDict()
	universal = ListenableDict()
	character = ListenableDict()
	time = MockTime()
	closed = False

	def __init__(self, *args, **kwargs):
		super().__init__(*args, **kwargs)
		self.turn = 0
		self._ready = True

	def __setattr__(self, key, value):
		if not hasattr(self, '_ready'):
			super().__setattr__(key, value)
			return
		self.send(self, key=key, value=value)
		super().__setattr__(key, value)

	def next_turn(self, *args, **kwargs):
		self.turn += 1
		kwargs['cb']('next_turn', 'trunk', self.turn, 0, ([], {}))

	def handle(self, *args, **kwargs):
		return {'a': 'b'}

	def commit(self):
		pass


class ELiDEAppTest(GraphicUnitTest):

	def setUp(self):
		super(ELiDEAppTest, self).setUp()
		self.prefix = mkdtemp()
		self.old_argv = sys.argv.copy()
		sys.argv = ['python', '-m', 'ELiDE', self.prefix]
		self.app = ELiDEApp()
		self.app.config = ConfigParser(None)
		self.app.build_config(self.app.config)

	def tearDown(self, fake=False):
		EventLoop.idle()  # let any triggered events fire
		super().tearDown(fake=fake)
		try:
			self.app.stop()
		except RedundantProcessError:
			pass
		shutil.rmtree(self.prefix)
		sys.argv = self.old_argv



# This file is part of ELiDE, frontend to LiSE, a framework for life simulation games.
# Copyright (c) Zachary Spector, public@zacharyspector.com
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, version 3.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
from kivy.clock import Clock
from kivy.uix.recycleview.layout import LayoutSelectionBehavior
from kivy.uix.recycleboxlayout import RecycleBoxLayout
from kivy.uix.behaviors import FocusBehavior
from functools import partial
from math import sin, cos, atan, pi

ninety = pi / 2
"""pi / 2"""

fortyfive = pi / 4
"""pi / 4"""


class SelectableRecycleBoxLayout(FocusBehavior, LayoutSelectionBehavior,
									RecycleBoxLayout):
	pass


class trigger(object):
	"""Make a trigger from a method.

    Decorate a method with this and it will become a trigger. Supply a
    numeric parameter to set a timeout.

    Not suitable for methods that expect any arguments other than
    ``dt``. However you should make your method accept ``*args`` for
    compatibility.

    """

	def __init__(self, func_or_timeout):
		if callable(func_or_timeout):
			self.func = func_or_timeout
			self.timeout = 0
		else:
			self.func = None
			self.timeout = func_or_timeout

	def __call__(self, func):
		self.func = func
		return self

	def __get__(self, instance, owner=None):
		if instance is None:
			# EventDispatcher iterates over its attributes before it
			# instantiates.  Don't try making any trigger in that
			# case.
			return
		retval = Clock.create_trigger(partial(self.func, instance),
										self.timeout)
		setattr(instance, self.func.__name__, retval)
		return retval


def try_load(loader, obj):
	"""Return the JSON interpretation the object if possible, or just the
    object otherwise.

    """
	try:
		return loader(obj)
	except (TypeError, ValueError):
		return obj


def dummynum(character, name):
	"""Count how many nodes there already are in the character whose name
    starts the same.

    """
	num = 0
	for nodename in character.node:
		nodename = str(nodename)
		if not nodename.startswith(name):
			continue
		try:
			nodenum = int(nodename.lstrip(name))
		except ValueError:
			continue
		num = max((nodenum, num))
	return num


def get_thin_rect_vertices(ox, oy, dx, dy, r):
	"""Given the starting point, ending point, and width, return a list of
    vertex coordinates at the corners of the line segment
    (really a thin rectangle).

    """
	if ox < dx:
		leftx = ox
		rightx = dx
		xco = 1
	elif ox > dx:
		leftx = ox * -1
		rightx = dx * -1
		xco = -1
	else:
		return [ox - r, oy, ox + r, oy, ox + r, dy, ox - r, dy]
	if oy < dy:
		boty = oy
		topy = dy
		yco = 1
	elif oy > dy:
		boty = oy * -1
		topy = dy * -1
		yco = -1
	else:
		return [ox, oy - r, dx, oy - r, dx, oy + r, ox, oy + r]

	rise = topy - boty
	run = rightx - leftx
	theta = atan(rise / run)
	theta_prime = ninety - theta
	xoff = cos(theta_prime) * r
	yoff = sin(theta_prime) * r
	x1 = leftx + xoff
	y1 = boty - yoff
	x2 = rightx + xoff
	y2 = topy - yoff
	x3 = rightx - xoff
	y3 = topy + yoff
	x4 = leftx - xoff
	y4 = boty + yoff
	return [
		x1 * xco, y1 * yco, x2 * xco, y2 * yco, x3 * xco, y3 * yco, x4 * xco,
		y4 * yco
	]



# This file is part of ELiDE, frontend to LiSE, a framework for life simulation games.
# Copyright (c) Zachary Spector, public@zacharyspector.com
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, version 3.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""Widget representing things that move about from place to place."""
from kivy.properties import (AliasProperty, ObjectProperty, NumericProperty,
								ReferenceListProperty)
from .pawnspot import GraphPawnSpot
from ..pawn import PawnBehavior


class Pawn(PawnBehavior, GraphPawnSpot):
	"""A token to represent a :class:`Thing`.

    :class:`Thing` is the LiSE class to represent items that are
    located in some :class:`Place` or other. Accordingly,
    :class:`Pawn`'s coordinates are never set directly; they are
    instead derived from the location of the :class:`Thing`
    represented. That means a :class:`Pawn` will appear next to the
    :class:`Spot` representing the :class:`Place` that its
    :class:`Thing` is in. The exception is if the :class:`Thing` is
    currently moving from its current :class:`Place` to another one,
    in which case the :class:`Pawn` will appear some distance along
    the :class:`Arrow` that represents the :class:`Portal` it's moving
    through.

    """

	def _get_location_wid(self):
		return self.board.spot[self.loc_name]

	def __repr__(self):
		"""Give my ``thing``'s name and its location's name."""
		return '<{}-in-{} at {}>'.format(self.name, self.loc_name, id(self))



# This file is part of ELiDE, frontend to LiSE, a framework for life simulation games.
# Copyright (c) Zachary Spector, public@zacharyspector.com
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, version 3.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""The big layout that you view all of ELiDE through.

Handles touch, selection, and time control. Contains a graph, a stat
grid, the time control panel, and the menu.

"""
from functools import partial
from ast import literal_eval

from kivy.app import App
from kivy.clock import mainthread
from kivy.factory import Factory
from kivy.lang import Builder
from kivy.uix.togglebutton import ToggleButton
from kivy.uix.button import Button
from kivy.uix.boxlayout import BoxLayout
from kivy.uix.floatlayout import FloatLayout
from kivy.uix.scrollview import ScrollView
from kivy.uix.screenmanager import Screen
from kivy.clock import Clock
from kivy.logger import Logger
from kivy.properties import (BooleanProperty, BoundedNumericProperty,
								DictProperty, ListProperty, NumericProperty,
								ObjectProperty, ReferenceListProperty,
								StringProperty)
from .stepper import RuleStepper
from .charmenu import CharMenu
from .graph.board import GraphBoardView
from .grid.board import GridBoardView
from .calendar import Agenda
from .util import dummynum, trigger

Factory.register('CharMenu', cls=CharMenu)


class KvLayout(FloatLayout):
	pass


class StatListPanel(BoxLayout):
	"""A panel that displays a simple two-column grid showing the stats of
    the selected entity, defaulting to those of the character being
    viewed.

    Has a button on the bottom to open the StatWindow in which
    to add and delete stats, or to change the way they are displayed
    in the StatListPanel.

    """
	selection_name = StringProperty()
	button_text = StringProperty('Configure stats')
	cfgstatbut = ObjectProperty()
	statlist = ObjectProperty()
	engine = ObjectProperty()
	proxy = ObjectProperty()
	toggle_stat_cfg = ObjectProperty()
	toggle_gridview = ObjectProperty()
	toggle_calendar = ObjectProperty()
	toggle_timestream = ObjectProperty()

	def on_proxy(self, *args):
		if hasattr(self.proxy, 'name'):
			self.selection_name = str(self.proxy.name)

	def set_value(self, k, v):
		if v is None:
			del self.proxy[k]
		else:
			try:
				vv = literal_eval(v)
			except (TypeError, ValueError):
				vv = v
			self.proxy[k] = vv


class SimulateButton(ToggleButton):
	play_arrow_left = NumericProperty()
	play_arrow_right = NumericProperty()
	play_arrow_points = ListProperty([0] * 6)
	graphics_top = NumericProperty()
	graphics_bot = NumericProperty()
	graphics_center_y = NumericProperty()

	def on_state(self, *args):
		app = App.get_running_app()
		app.edit_locked = self.state == 'down'


class OneTurnButton(Button):
	graphics_top = NumericProperty()
	graphics_bot = NumericProperty()
	graphics_center_y = NumericProperty()
	step_arrow_left = NumericProperty()
	step_center_x = NumericProperty()
	step_bar_right = NumericProperty()
	step_arrow_points = ListProperty([0] * 6)
	step_rect_points = ListProperty([0] * 8)
	screen = ObjectProperty()

	def on_release(self):
		App.get_running_app().edit_locked = True
		self.screen.next_turn(cb=self._release_edit_lock)

	@staticmethod
	def _release_edit_lock(*args):
		App.get_running_app().edit_locked = False


class TimePanel(BoxLayout):
	"""A panel that lets you to start and stop the game, or browse through
    its history.

    There's a "simulate" button, which is toggleable. When toggled on, the
    simulation will continue to run until it's toggled off
    again. Next to this is a "1 turn" button, which will simulate
    exactly one turn and stop. And there are two text fields in which
    you can manually enter a Branch and Tick to go to. Moving through
    time this way doesn't simulate anything--you'll only see what
    happened as a result of "simulate," "1 turn," or some other way
    the LiSE rules engine has been made to run.

    """
	screen = ObjectProperty()
	buttons_font_size = NumericProperty(18)
	disable_one_turn = BooleanProperty()

	def set_branch(self, *args):
		branch = self.ids.branchfield.text
		self.ids.branchfield.text = ''
		self.screen.app.time_travel(branch, self.screen.app.turn)
		self.screen.charmenu._switch_to_menu()

	def set_turn(self, *args):
		turn = int(self.ids.turnfield.text)
		self.ids.turnfield.text = ''
		self.screen.app.time_travel(self.screen.app.branch, turn)
		self.screen.charmenu._switch_to_menu()

	def set_tick(self, *args):
		tick = int(self.ids.tickfield.text)
		self.ids.tickfield.text = ''
		self.screen.app.time_travel(self.screen.app.branch,
									self.screen.app.turn, tick)
		self.screen.charmenu._switch_to_menu()

	@mainthread
	def _upd_branch_hint(self, app, *args):
		self.ids.branchfield.hint_text = app.branch

	@mainthread
	def _upd_turn_hint(self, app, *args):
		self.ids.turnfield.hint_text = str(app.turn)

	@mainthread
	def _upd_tick_hint(self, app, *args):
		self.ids.tickfield.hint_text = str(app.tick)

	def on_screen(self, *args):
		if not all(field in self.ids
					for field in ('branchfield', 'turnfield', 'tickfield')):
			Clock.schedule_once(self.on_screen, 0)
			return
		self.ids.branchfield.hint_text = self.screen.app.branch
		self.ids.turnfield.hint_text = str(self.screen.app.turn)
		self.ids.tickfield.hint_text = str(self.screen.app.tick)
		self.screen.app.bind(branch=self._upd_branch_hint,
								turn=self._upd_turn_hint,
								tick=self._upd_tick_hint)


class MainScreen(Screen):
	"""A master layout that contains one graph and some menus.

    This contains three elements: a scrollview (containing the graph),
    a menu, and the time control panel. This class has some support methods
    for handling interactions with the menu and the character sheet,
    but if neither of those happen, the scrollview handles touches on its
    own.

    """
	manager = ObjectProperty()
	graphboards = DictProperty()
	gridboards = DictProperty()
	boardview = ObjectProperty()
	mainview = ObjectProperty()
	charmenu = ObjectProperty()
	statlist = ObjectProperty()
	statpanel = ObjectProperty()
	timepanel = ObjectProperty()
	kv = StringProperty()
	use_kv = BooleanProperty()
	play_speed = NumericProperty()
	playbut = ObjectProperty()
	portaladdbut = ObjectProperty()
	portaldirbut = ObjectProperty()
	dummyplace = ObjectProperty()
	dummything = ObjectProperty()
	dummies = ReferenceListProperty(dummyplace, dummything)
	dialoglayout = ObjectProperty()
	visible = BooleanProperty()
	_touch = ObjectProperty(None, allownone=True)
	rules_per_frame = BoundedNumericProperty(10, min=1)
	tmp_block = BooleanProperty(False)

	@property
	def app(self):
		return App.get_running_app()

	def _update_adding_portal(self, *args):
		self.boardview.adding_portal = self.charmenu.portaladdbut.state == 'down'

	def _update_board(self, *args):
		self.boardview.board = self.graphboards[self.app.character_name]
		self.gridview.board = self.gridboards[self.app.character_name]

	def on_mainview(self, *args):
		if None in (self.statpanel, self.charmenu, self.app) or None in (
			self.app.character_name, self.charmenu.portaladdbut
		) or self.app.character_name not in self.graphboards:
			Clock.schedule_once(self.on_mainview, 0)
			return
		self.boardview = GraphBoardView(
			scale_min=0.2,
			scale_max=4.0,
			size=self.mainview.size,
			pos=self.mainview.pos,
			board=self.graphboards[self.app.character_name],
			adding_portal=self.charmenu.portaladdbut.state == 'down')
		self.mainview.bind(size=self.boardview.setter('size'),
							pos=self.boardview.setter('pos'))
		self.charmenu.portaladdbut.bind(state=self._update_adding_portal)
		self.app.bind(character_name=self._update_board)
		self.calendar = Agenda(update_mode='present')
		self.calendar_view = ScrollView(size=self.mainview.size,
										pos=self.mainview.pos)
		self.gridview = GridBoardView(
			scale_min=0.2,
			scale_max=4.0,
			size=self.mainview.size,
			pos=self.mainview.pos,
			board=self.gridboards[self.app.character_name])
		self.mainview.bind(
			size=self.calendar_view.setter('size'),
			pos=self.calendar_view.setter('pos'),
		)
		self.mainview.bind(size=self.gridview.setter('size'),
							pos=self.gridview.setter('pos'))
		self.calendar_view.add_widget(self.calendar)
		self.mainview.add_widget(self.boardview)

	def on_statpanel(self, *args):
		if not self.app:
			Clock.schedule_once(self.on_statpanel, 0)
			return
		self.app.bind(selected_proxy=self._update_statlist,
						branch=self._update_statlist,
						turn=self._update_statlist,
						tick=self._update_statlist)

	@trigger
	def _update_statlist(self, *args):
		if not self.app or not self.app.engine:
			return
		if not self.app.selected_proxy:
			self._update_statlist()
			return
		self.app.update_calendar(self.statpanel.statlist,
									past_turns=0,
									future_turns=0)

	def pull_visibility(self, *args):
		self.visible = self.manager.current == 'main'

	def on_manager(self, *args):
		self.pull_visibility()
		self.manager.bind(current=self.pull_visibility)

	def on_play_speed(self, *args):
		"""Change the interval at which ``self.play`` is called to match my
        current ``play_speed``.

        """
		if hasattr(self, '_play_scheduled'):
			Clock.unschedule(self._play_scheduled)
		self._play_scheduled = Clock.schedule_interval(self.play,
														1.0 / self.play_speed)

	def remake_display(self, *args):
		"""Remake any affected widgets after a change in my ``kv``.

        """
		Builder.load_string(self.kv)
		if hasattr(self, '_kv_layout'):
			self.remove_widget(self._kv_layout)
			del self._kv_layout
		self._kv_layout = KvLayout()
		self.add_widget(self._kv_layout)

	_trigger_remake_display = trigger(remake_display)

	def on_touch_down(self, touch):
		if self.visible:
			touch.grab(self)
		for interceptor in (self.timepanel, self.charmenu, self.statpanel,
							self.dummyplace, self.dummything):
			if interceptor.collide_point(*touch.pos):
				interceptor.dispatch('on_touch_down', touch)
				self.boardview.keep_selection = \
                             self.gridview.keep_selection = True
				return True
		if self.dialoglayout.dispatch('on_touch_down', touch):
			return True
		return self.mainview.dispatch('on_touch_down', touch)

	def on_touch_up(self, touch):
		if self.timepanel.collide_point(*touch.pos):
			return self.timepanel.dispatch('on_touch_up', touch)
		elif self.charmenu.collide_point(*touch.pos):
			return self.charmenu.dispatch('on_touch_up', touch)
		elif self.statpanel.collide_point(*touch.pos):
			return self.statpanel.dispatch('on_touch_up', touch)
		return self.mainview.dispatch('on_touch_up', touch)

	def on_dummies(self, *args):
		"""Give the dummies numbers such that, when appended to their names,
        they give a unique name for the resulting new
        :class:`graph.Pawn` or :class:`graph.Spot`.

        """
		if not self.app.character:
			Clock.schedule_once(self.on_dummies, 0)
			return

		def renum_dummy(dummy, *args):
			dummy.num = dummynum(self.app.character, dummy.prefix) + 1

		for dummy in self.dummies:
			if dummy is None or hasattr(dummy, '_numbered'):
				continue
			if dummy == self.dummything:
				self.app.pawncfg.bind(imgpaths=self._propagate_thing_paths)
			if dummy == self.dummyplace:
				self.app.spotcfg.bind(imgpaths=self._propagate_place_paths)
			dummy.num = dummynum(self.app.character, dummy.prefix) + 1
			Logger.debug("MainScreen: dummy #{}".format(dummy.num))
			dummy.bind(prefix=partial(renum_dummy, dummy))
			dummy._numbered = True

	def _propagate_thing_paths(self, *args):
		# horrible hack
		self.dummything.paths = self.app.pawncfg.imgpaths

	def _propagate_place_paths(self, *args):
		# horrible hack
		self.dummyplace.paths = self.app.spotcfg.imgpaths

	def _update_from_time_travel(self, command, branch, turn, tick, result,
									**kwargs):
		self._update_from_delta(command, branch, turn, tick, result[-1])

	def _update_from_delta(self, cmd, branch, turn, tick, delta, **kwargs):
		self.app.branch = branch
		self.app.turn = turn
		self.app.tick = tick
		chardelta = delta.get(self.boardview.board.character.name, {})
		for unwanted in ('character_rulebook', 'avatar_rulebook',
							'character_thing_rulebook',
							'character_place_rulebook',
							'character_portal_rulebook'):
			if unwanted in chardelta:
				del chardelta[unwanted]
		self.boardview.board.trigger_update_from_delta(chardelta)
		self.gridview.board.trigger_update_from_delta(chardelta)
		self.statpanel.statlist.mirror = dict(self.app.selected_proxy)

	def play(self, *args):
		"""If the 'play' button is pressed, advance a turn.

        If you want to disable this, set ``engine.universal['block'] = True``

        """
		if self.playbut.state == 'normal' or not hasattr(self.app, 'engine') or \
                self.app.engine is None or self.app.engine.closed:
			return
		self.next_turn()

	def _update_from_next_turn(self,
								command,
								branch,
								turn,
								tick,
								result,
								cb=None):
		todo, deltas = result
		if isinstance(todo, list):
			self.dialoglayout.todo = todo
			self.dialoglayout.idx = 0
		self._update_from_delta(command, branch, turn, tick, deltas)
		self.dialoglayout.advance_dialog()
		self.tmp_block = False
		if cb is not None:
			cb(command, branch, turn, tick, result)

	def next_turn(self, cb=None, *args):
		"""Advance time by one turn, if it's not blocked.

        Block time by setting ``engine.universal['block'] = True``"""
		if self.tmp_block:
			return
		eng = self.app.engine
		dial = self.dialoglayout
		if eng.universal.get('block'):
			Logger.info(
				"MainScreen: next_turn blocked, delete universal['block'] to unblock"
			)
			return
		if dial.idx < len(dial.todo):
			Logger.info(
				"MainScreen: not advancing time while there's a dialog")
			return
		self.tmp_block = True
		eng.next_turn(cb=partial(self._update_from_next_turn, cb=cb))
		self.ids.charmenu._switch_to_menu()

	def switch_to_calendar(self, *args):
		self.app.update_calendar(self.calendar)
		self.mainview.clear_widgets()
		self.mainview.add_widget(self.calendar_view)

	def switch_to_boardview(self, *args):
		self.mainview.clear_widgets()
		self.app.engine.handle('apply_choices',
								choices=[self.calendar.get_track()])
		self.mainview.add_widget(self.boardview)

	def toggle_gridview(self, *args):
		if self.gridview in self.mainview.children:
			self.mainview.clear_widgets()
			self.mainview.add_widget(self.boardview)
		else:
			self.mainview.clear_widgets()
			self.mainview.add_widget(self.gridview)

	def toggle_calendar(self, *args):
		# TODO decide how to handle switching between >2 view types
		if self.boardview in self.mainview.children:
			self.switch_to_calendar()
		else:
			self.switch_to_boardview()

	@trigger
	def toggle_timestream(self, *args):
		if self.manager.current != 'timestream':
			self.manager.current = 'timestream'
		else:
			self.manager.current = 'main'


class CharMenuContainer(BoxLayout):
	screen = ObjectProperty()
	dummyplace = ObjectProperty()
	dummything = ObjectProperty()
	portaladdbut = ObjectProperty()

	def __init__(self, **kwargs):
		super(CharMenuContainer, self).__init__(**kwargs)
		self.charmenu = CharMenu(screen=self.screen, size_hint_y=0.9)
		self.bind(screen=self.charmenu.setter('screen'))
		self.dummyplace = self.charmenu.dummyplace
		self.charmenu.bind(dummyplace=self.setter('dummyplace'))
		self.dummything = self.charmenu.dummything
		self.charmenu.bind(dummything=self.setter('dummything'))
		self.portaladdbut = self.charmenu.portaladdbut
		self.charmenu.bind(portaladdbut=self.setter('portaladdbut'))
		self.stepper = RuleStepper(size_hint_y=0.9)
		self.button = Button(on_release=self._toggle,
								text='Rule\nstepper',
								size_hint_y=0.1)

	def on_parent(self, *args):
		if not self.screen or not hasattr(self, 'charmenu') or not hasattr(
			self, 'stepper') or not hasattr(self, 'button'):
			Clock.schedule_once(self.on_parent, 0)
			return
		self.add_widget(self.charmenu)
		self.add_widget(self.button)

	@trigger
	def _toggle(self, *args):
		if self.charmenu in self.children:
			engine = self.screen.app.engine
			self.clear_widgets()
			self.stepper.from_rules_handled_turn(
				engine.handle('rules_handled_turn'))
			self.add_widget(self.stepper)
			self.button.text = 'Menu'
		else:
			self.clear_widgets()
			self.add_widget(self.charmenu)
			self.button.text = 'Rule stepper'
		self.add_widget(self.button)

	@trigger
	def _switch_to_menu(self, *args):
		if self.charmenu not in self.children:
			self.clear_widgets()
			self.add_widget(self.charmenu)
			self.button.text = 'Rule stepper'
			self.add_widget(self.button)


Builder.load_string("""
#: import resource_find kivy.resources.resource_find
<StatListPanel>:
    orientation: 'vertical'
    cfgstatbut: cfgstatbut
    statlist: statlist
    id: statpanel
    proxy: app.selected_proxy
    Label:
        size_hint_y: 0.05
        text: root.selection_name
        bold: True
    Calendar:
        id: statlist
        size_hint_y: 0.8
        entity: root.proxy
        update_mode: 'present'
        disabled: app.edit_locked
    Button:
        id: gridviewbut
        size_hint_y: 0.05
        text: 'Toggle grid'
        on_release: root.toggle_gridview()
    Button:
        id: cfgstatbut
        size_hint_y: 0.05
        text: root.button_text
        on_release: root.toggle_stat_cfg()
    Button:
        id: timestreambut
        size_hint_y: 0.05
        text: 'Timestream'
        on_release: root.toggle_timestream()
<SimulateButton>:
    graphics_top: self.y + self.font_size + (self.height - self.font_size) * (3/4)
    graphics_bot: self.y + self.font_size + 3
    graphics_center_y: self.graphics_bot + (self.graphics_top - self.graphics_bot) / 2
    play_arrow_left: self.center_x - self.width / 6
    play_arrow_right: self.center_x + self.width / 6
    play_arrow_points: self.play_arrow_left, self.graphics_top, self.play_arrow_right, self.graphics_center_y, self.play_arrow_left, self.graphics_bot
    canvas:
        Triangle:
            points: root.play_arrow_points
        SmoothLine:
            points: root.play_arrow_points[:-2] + [root.play_arrow_points[-2]+1, root.play_arrow_points[-1]+1]
    Label:
        id: playlabel
        font_size: root.font_size
        center_x: root.center_x
        y: root.y
        size: self.texture_size
        text: 'Simulate'
<OneTurnButton>:
    graphics_top: self.y + self.font_size + (self.height - self.font_size) * (3/4)
    graphics_bot: self.y + self.font_size + 3
    graphics_center_y: self.graphics_bot + (self.graphics_top - self.graphics_bot) / 2
    step_arrow_left: self.center_x - (self.width / 6)
    step_center_x: self.center_x + self.width / 6
    step_bar_right: self.center_x + self.width / 4
    step_arrow_points: self.step_arrow_left, self.graphics_top, self.step_center_x, self.graphics_center_y, self.step_arrow_left, self.graphics_bot
    step_rect_points: self.step_center_x, self.graphics_top, self.step_bar_right, self.graphics_top, self.step_bar_right, self.graphics_bot, self.step_center_x, self.graphics_bot
    canvas:
        Triangle:
            points: root.step_arrow_points
        Quad:
            points: root.step_rect_points 
        SmoothLine:
            points: root.step_arrow_points
        SmoothLine:
            points: root.step_rect_points 
    Label:
        font_size: root.font_size
        center_x: root.center_x
        y: root.y
        size: self.texture_size
        text: '1 turn'
<TimePanel>:
    orientation: 'vertical'
    playbut: playbut
    BoxLayout:
        size_hint_y: 0.4
        BoxLayout:
            orientation: 'vertical'
            Label:
                size_hint_y: 0.4
                text: 'Branch'
            MenuTextInput:
                id: branchfield
                set_value: root.set_branch
                hint_text: root.screen.app.branch if root.screen else ''
        BoxLayout:
            BoxLayout:
                orientation: 'vertical'
                Label:
                    size_hint_y: 0.4
                    text: 'Turn'
                MenuIntInput:
                    id: turnfield
                    set_value: root.set_turn
                    hint_text: str(root.screen.app.turn) if root.screen else ''
            BoxLayout:
                orientation: 'vertical'
                Label:
                    size_hint_y: 0.4
                    text: 'Tick'
                MenuIntInput:
                    id: tickfield
                    set_value: root.set_tick
                    hint_text: str(root.screen.app.tick) if root.screen else ''
    BoxLayout:
        size_hint_y: 0.6
        SimulateButton:
            id: playbut
            font_size: root.buttons_font_size
        OneTurnButton:
            id: stepbut
            font_size: root.buttons_font_size
            screen: root.screen
            disabled: root.disable_one_turn or app.edit_locked
<MainScreen>:
    name: 'main'
    dummyplace: charmenu.dummyplace
    dummything: charmenu.dummything
    mainview: mainview
    playbut: timepanel.playbut
    portaladdbut: charmenu.portaladdbut
    charmenu: charmenu
    statlist: statpanel.statlist
    statpanel: statpanel
    timepanel: timepanel
    dialoglayout: dialoglayout
    Widget:
        id: mainview
        x: statpanel.right
        y: 0
        size_hint: (None, None)
        width: charmenu.x - statpanel.right
        height: root.height
    StatListPanel:
        id: statpanel
        engine: app.engine
        toggle_stat_cfg: app.statcfg.toggle
        toggle_calendar: root.toggle_calendar
        toggle_gridview: root.toggle_gridview
        pos_hint: {'left': 0, 'top': 1}
        size_hint: (0.25, 0.8)
        selection_name: app.selected_proxy_name
        toggle_timestream: root.toggle_timestream
    TimePanel:
        id: timepanel
        screen: root
        pos_hint: {'bot': 0}
        size_hint: (0.25, 0.2)
        disable_one_turn: root.tmp_block
    CharMenuContainer:
        id: charmenu
        orientation: 'vertical'
        screen: root
        pos_hint: {'right': 1, 'top': 1}
        size_hint: (0.1, 1)
    DialogLayout:
        id: dialoglayout
        engine: app.engine
        size_hint: None, None
        x: statpanel.right
        y: timepanel.top
        width: charmenu.x - statpanel.right
        height: root.height - timepanel.top
""")



# This file is part of ELiDE, frontend to LiSE, a framework for life simulation games.
# Copyright (c) Zachary Spector, public@zacharyspector.com
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, version 3.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""Grid of current values for some entity. Can be changed by the
user. Autoupdates when there's a change for any reason.

"""
from functools import partial
from kivy.properties import (
	BooleanProperty,
	DictProperty,
	ObjectProperty,
	StringProperty,
)
from kivy.logger import Logger
from kivy.clock import Clock
from kivy.uix.widget import Widget
from kivy.uix.boxlayout import BoxLayout
from kivy.uix.recycleview import RecycleView
from kivy.uix.togglebutton import ToggleButton
from kivy.uix.slider import Slider
from kivy.uix.textinput import TextInput
from kivy.uix.label import Label
from kivy.lang import Builder
from .util import trigger


class StatRowListItem(Widget):
	"""Mixin class for widgets that show a stat's current value.

    """
	key = ObjectProperty()
	"The stat I am about. Same as in the LiSE entity."
	value = ObjectProperty(None, allownone=True)
	"The present value of the stat."
	gett = ObjectProperty()
	"Getter function: takes stat, returns its present value."
	sett = ObjectProperty()
	"Setter function: takes stat and value, sets stat=value."
	listen = ObjectProperty()
	"Function that takes a function and calls it when the stat changes."
	unlisten = ObjectProperty()
	"Function that takes a ``listen``ed function and stops it listening."
	config = DictProperty()
	"Dictionary of some parameters for how to present myself."

	def __init__(self, **kwargs):
		super().__init__(**kwargs)
		self.bind(value=self._bind_value)

	def _bind_value(self, *args):
		self.bind(value=self._push)
		self.unbind(value=self._bind_value)

	def on_listen(self, *args):
		self.listen(self._pull)
		self._pull()

	@trigger
	def _push(self, *args):
		self.sett(self.key, self.value)

	def _really_pull(self, *args):
		self.unbind(value=self._push)
		try:
			self.value = self.gett(self.key)
		except KeyError:
			Logger.info('StatRowListItem: {} deleted'.format(self.key))
		except TypeError:
			value = self.gett(self.key)
			Logger.warning(
				"StatRowListItem: couldn't set value {} because it is type {}".
				format(value, type(value)))
		self.bind(value=self._push)

	def _pull(self, *args, **kwargs):
		if hasattr(self, '_scheduled_pull'):
			Clock.unschedule(self._scheduled_pull)
		self._scheduled_pull = Clock.schedule_once(self._really_pull, 0)


class StatRowLabel(StatRowListItem, Label):
	"""Display the current value of a stat as text."""


class StatRowTextInput(StatRowListItem, TextInput):
	"""Display the current value of a stat and accept text input to change it."""

	def __init__(self, **kwargs):
		kwargs['multiline'] = False
		super().__init__(**kwargs)

		def lost_focus(self, *args):
			if not self.focus:
				self.upd_value()

		self.bind(on_enter=self.upd_value,
					on_text_validate=self.upd_value,
					on_focus=lost_focus)

	@trigger
	def upd_value(self, *args):
		if self.text not in ('', self.value):
			self.value = self.text
		self.text = ''


class StatRowToggleButton(StatRowListItem, ToggleButton):
	"""Display the current value of a stat, 0 or 1, and let the user press a button to flip it."""
	true_text = StringProperty('1')
	"""String to display when the stat is true."""
	false_text = StringProperty('0')
	"""String to display when the stat is false."""

	def on_touch_up(self, *args):
		if self.parent is None:
			return
		if self.state == 'normal' and self.value != 0:
			self.value = 0
		elif self.state == 'down' and self.value != 1:
			self.value = 1


class StatRowSlider(StatRowListItem, Slider):
	"""Display the current value of a numeric stat and let the user slide it."""
	need_set = BooleanProperty(False)
	"""Internal. Usually False, becomes True briefly when the value has changed."""

	def __init__(self, **kwargs):
		self.value = kwargs['value']
		self.min = kwargs['min']
		self.max = kwargs['max']
		super().__init__(**kwargs)

	def _bind_value(self, *args):
		pass

	def _really_pull(self, *args):
		try:
			self.value = self.gett(self.key)
		except KeyError:
			Logger.info('StatRowListItem: {} deleted'.format(self.key))
		except TypeError:
			value = self.gett(self.key)
			Logger.warning(
				"StatRowListItem: couldn't set value {} because it is type {}".
				format(value, type(value)))

	def on_listen(self, *args):
		self.listen(self._pull)
		self._pull()

	def on_value(self, *args):
		self.need_set = True

	def on_touch_up(self, *args):
		if self.need_set:
			self._push()
			self.need_set = False


class StatRowListItemContainer(BoxLayout):
	"""The name of a stat followed by a widget representing its value.

    The widget can be

    * :class:`StatRowLabel`
    * :class:`StatRowTextInput`
    * :class:`StatRowToggleButton`
    * :class:`StatRowSlider`

    """
	key = ObjectProperty()
	"""The name of the stat"""
	reg = ObjectProperty()
	unreg = ObjectProperty()
	gett = ObjectProperty()
	"""Function to get the current value of stats, taking the stat name as its argument"""
	sett = ObjectProperty()
	"""Function to set the current value of stats, taking args (key, value)"""
	listen = ObjectProperty()
	"""Function to register a listener to a LiSE entity"""
	unlisten = ObjectProperty()
	"""Function to unregister a listener from a LiSE entity"""
	config = DictProperty()
	"""Dictionary describing the configuration of this stat's widget.
    
    The key 'control' has the widget type as its value, which may be
    
    * 'readout'
    * 'textinput'
    * 'togglebutton'
    * 'slider'
    
    Other keys are specific to one widget type or another.
    
    """

	def set_value(self, *args):
		"""Use my ``sett`` function to set my stat (``key``) to my new ``value``.

        This doesn't need arguments, but accepts any positional arguments provided,
        so that you can use this in kvlang

        """
		self.sett(self.key, self.value)

	def __init__(self, **kwargs):
		kwargs.setdefault('orientation', 'vertical')
		super().__init__(**kwargs)
		self.remake()

	def on_key(self, *args):
		self.remake()

	def on_config(self, *args):
		self.remake()

	def on_parent(self, *args):
		self.remake()

	def remake(self, *args):
		"""Replace any existing child widget with the one described by my ``config``.

        This doesn't need arguments, but accepts any positional arguments provided,
        so that you can use this in kvlang

        """
		if not self.config:
			return
		if not all(
			(self.key, self.gett, self.sett, self.listen, self.unlisten)):
			Clock.schedule_once(self.remake, 0)
			return
		if not hasattr(self, 'label'):
			self.label = Label(text=str(self.key))

			def updlabel(*args):
				self.label.text = str(self.key)

			self.bind(key=updlabel)
			self.add_widget(self.label)
		if hasattr(self, 'wid'):
			self.remove_widget(self.wid)
			del self.wid
		control = self.config['control']
		widkwargs = {
			'key': self.key,
			'gett': self.gett,
			'sett': self.sett,
			'listen': self.listen,
			'unlisten': self.unlisten
		}
		if control == 'slider':
			cls = StatRowSlider
			try:
				widkwargs['value'] = float(self.gett(self.key))
				widkwargs['min'] = float(self.config['min'])
				widkwargs['max'] = float(self.config['max'])
			except (KeyError, ValueError):
				return
		elif control == 'togglebutton':
			cls = StatRowToggleButton
			try:
				widkwargs['true_text'] = self.config['true_text']
				widkwargs['false_text'] = self.config['false_text']
			except KeyError:
				return
		elif control == 'textinput':
			cls = StatRowTextInput
		else:
			cls = StatRowLabel
		self.wid = cls(**widkwargs)
		self.bind(key=self.wid.setter('key'),
					gett=self.wid.setter('gett'),
					sett=self.wid.setter('sett'),
					listen=self.wid.setter('listen'),
					unlisten=self.wid.setter('unlisten'))
		if control == 'slider':
			self.unbind(config=self._toggle_update_config)
			self.bind(config=self._slider_update_config)
		elif control == 'togglebutton':
			self.unbind(config=self._slider_update_config)
			self.bind(config=self._toggle_update_config)
		else:
			self.unbind(config=self._slider_update_config)
			self.unbind(config=self._toggle_update_config)
		self.add_widget(self.wid)

	@trigger
	def _slider_update_config(self, *args):
		self.wid.min = self.config['min']
		self.wid.max = self.config['max']

	@trigger
	def _toggle_update_config(self, *args):
		self.wid.true_text = self.config['true_text']
		self.wid.false_text = self.config['false_text']


default_cfg = {
	'control': 'readout',
	'true_text': '1',
	'false_text': '0',
	'min': 0.0,
	'max': 1.0
}


class BaseStatListView(RecycleView):
	"""Base class for widgets showing lists of stats and their values"""
	proxy = ObjectProperty()
	"""A proxy object representing a LiSE entity"""
	engine = ObjectProperty()
	"""A :class:`LiSE.proxy.EngineProxy` object"""
	app = ObjectProperty()
	"""The Kivy app object"""
	_scheduled_set_value = DictProperty()

	def __init__(self, **kwargs):
		self._listeners = {}
		super().__init__(**kwargs)

	def on_proxy(self, *args):
		self.proxy.connect(self._trigger_upd_data, weak=False)
		self._trigger_upd_data()

	def del_key(self, k):
		"""Delete the key and any configuration for it"""
		if k not in self.proxy:
			raise KeyError
		del self.proxy[k]
		if '_config' in self.proxy and k in self.proxy['_config']:
			del self.proxy['_config'][k]

	def set_value(self, k, v):
		"""Set a value on the proxy, parsing it to a useful datatype if possible"""
		from ast import literal_eval
		if self.engine is None or self.proxy is None:
			self._trigger_set_value(k, v)
			return
		if v is None:
			del self.proxy[k]
		else:
			try:
				vv = literal_eval(v)
			except (TypeError, ValueError):
				vv = v
			self.proxy[k] = vv
		if (k, v) in self._scheduled_set_value:
			del self._scheduled_set_value[k, v]

	def _trigger_set_value(self, k, v, *args):
		todo = partial(self.set_value, k, v)
		if (k, v) in self._scheduled_set_value:
			Clock.unschedule(self._scheduled_set_value[k, v])
		self._scheduled_set_value[k, v] = Clock.schedule_once(todo, 0)

	def init_config(self, key):
		"""Set the configuration for the key to something that will always work"""
		self.proxy['_config'].setdefault(key, default_cfg)

	def set_config(self, key, option, value):
		"""Set a configuration option for a key"""
		if '_config' not in self.proxy:
			newopt = dict(default_cfg)
			newopt[option] = value
			self.proxy['_config'] = {key: newopt}
		else:
			if key in self.proxy['_config']:
				self.proxy['_config'][key][option] = value
			else:
				newopt = dict(default_cfg)
				newopt[option] = value
				self.proxy['_config'][key] = newopt

	def set_configs(self, key, d):
		"""Set the whole configuration for a key"""
		if '_config' in self.proxy:
			self.proxy['_config'][key] = d
		else:
			self.proxy['_config'] = {key: d}

	def iter_data(self):
		"""Iterate over key-value pairs that are really meant to be displayed"""
		invalid = {'character', 'name', 'location', 'rulebooks'}
		for (k, v) in self.proxy.items():
			if (not (isinstance(k, str) and k[0] == '_') and k not in invalid):
				yield k, v

	def munge(self, k, v):
		"""Turn a key and value into a dictionary describing a widget to show"""
		if '_config' in self.proxy and k in self.proxy['_config']:
			config = self.proxy['_config'][k].unwrap()
		else:
			config = default_cfg
		return {
			'key': k,
			'reg': self._reg_widget,
			'unreg': self._unreg_widget,
			'gett': self.proxy.__getitem__,
			'sett': self.set_value,
			'listen': self.proxy.connect,
			'unlisten': self.proxy.disconnect,
			'config': config
		}

	def upd_data(self, *args):
		"""Update to match new entity data"""
		data = [self.munge(k, v) for k, v in self.iter_data()]
		self.data = sorted(data, key=lambda d: d['key'])

	def _trigger_upd_data(self, *args, **kwargs):
		if hasattr(self, '_scheduled_upd_data'):
			Clock.unschedule(self._scheduled_upd_data)
		self._scheduled_upd_data = Clock.schedule_once(self.upd_data, 0)

	def _reg_widget(self, w, *args):
		if not self.proxy:
			Clock.schedule_once(partial(self._reg_widget, w), 0)
			return

		def listen(*args):
			if w.key not in self.proxy:
				return
			if w.value != self.proxy[w.key]:
				w.value = self.proxy[w.key]

		self._listeners[w.key] = listen
		self.proxy.connect(listen)

	def _unreg_widget(self, w):
		if w.key in self._listeners:
			self.proxy.disconnect(self._listeners[w.key])


class StatListView(BaseStatListView):
	"""The type of StatListView that shows only stats and their values"""


Builder.load_string("""
<StatRowListItem>:
    height: 30
<StatRowLabel>:
    text: str(self.value)
<StatRowTextInput>:
    hint_text: str(self.value)
    multiline: False
<StatRowToggleButton>:
    text: self.true_text if self.value else self.false_text
    state: 'down' if self.value else 'normal'
<StatRowSlider>:
    Label:
        center_x: root.center_x
        y: root.center_y
        text: str(root.value)
        size: self.texture_size
<StatListView>:
    viewclass: 'StatRowListItemContainer'
    app: app
    RecycleBoxLayout:
        default_size: None, dp(56)
        default_size_hint: 1, None
        size_hint_y: None
        height: self.minimum_height
        orientation: 'vertical'
""")



from math import floor, ceil
from array import array

# Copyright (c) Kivy Team and other contributors
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.


class Collide2DPoly(object):
	''' Collide2DPoly checks whether a point is within a polygon defined by a
    list of corner points.

    Based on http://alienryderflex.com/polygon/

    For example, a simple triangle::

        >>> collider = Collide2DPoly([10., 10., 20., 30., 30., 10.],
        ... cache=True)
        >>> (0.0, 0.0) in collider
        False
        >>> (20.0, 20.0) in collider
        True

    The constructor takes a list of x,y points in the form of [x1,y1,x2,y2...]
    as the points argument. These points define the corners of the
    polygon. The boundary is linearly interpolated between each set of points.
    The x, and y values must be floating points.
    The cache argument, if True, will calculate membership for all the points
    so when collide_point is called it'll just be a table lookup.

    This pure Python version was ported from the kivy.garden.collider
    package. It is available under the MIT license.
    '''

	def __init__(self, points, cache=False, **kwargs):
		length = len(points)
		if length % 2:
			raise IndexError()
		if length < 4:
			self.points = None
			return
		count = length // 2
		self.count = count
		self.points = points = array('d', points)
		self.constant = constant = array('d', [0.0] * count)
		self.multiple = multiple = array('d', [0.0] * count)

		self.min_x = min(points[0::2])
		self.max_x = max(points[0::2])
		self.min_y = min(points[1::2])
		self.max_y = max(points[1::2])
		min_x = floor(self.min_x)
		min_y = floor(self.min_y)
		j = count - 1
		if cache:
			for i in range(count):
				points[2 * i] -= min_x
				points[2 * i + 1] -= min_y

		for i in range(count):
			i_x = i * 2
			i_y = i_x + 1
			j_x = j * 2
			j_y = j_x + 1
			if points[j_y] == points[i_y]:
				constant[i] = points[i_x]
				multiple[i] = 0.
			else:
				constant[i] = (points[i_x] - points[i_y] * points[j_x] /
								(points[j_y] - points[i_y]) +
								points[i_y] * points[i_x] /
								(points[j_y] - points[i_y]))
				multiple[i] = ((points[j_x] - points[i_x]) /
								(points[j_y] - points[i_y]))
			j = i
		if cache:
			width = int(ceil(self.max_x) - min_x + 1.)
			self.width = width
			height = int(ceil(self.max_y) - min_y + 1.)
			self.space = space = []
			for y in range(height):
				for x in range(width):
					j = count - 1
					odd = 0
					for i in range(count):
						i_y = i * 2 + 1
						j_y = j * 2 + 1
						if (points[i_y] < y and points[j_y] >= y
							or points[j_y] < y and points[i_y] >= y):
							odd ^= y * multiple[i] + constant[i] < x
						j = i
					space[y * width + x] = odd

	def collide_point(self, x, y):
		points = self.points
		if not points or not (self.min_x <= x <= self.max_x
								and self.min_y <= y <= self.max_y):
			return False
		if hasattr(self, 'space'):
			y -= floor(self.min_y)
			x -= floor(self.min_x)
			return self.space[int(y) * self.width + int(x)]

		j = self.count - 1
		odd = 0
		multiple = self.multiple
		constant = self.constant
		for i in range(self.count):
			i_y = i * 2 + 1
			j_y = j * 2 + 1
			if (points[i_y] < y and points[j_y] >= y
				or points[j_y] < y and points[i_y] >= y):
				odd ^= y * multiple[i] + constant[i] < x
			j = i
		return odd

	def __contains__(self, point):
		return self.collide_point(*point)



# This file is part of ELiDE, frontend to LiSE, a framework for life simulation games.
# Copyright (c) Zachary Spector, public@zacharyspector.com
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, version 3.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""Widget that looks like a trading card, and a layout within which it
can be dragged and dropped to some particular position within stacks
of other cards.

"""

from kivy.clock import Clock
from kivy.lang import Builder
from kivy.logger import Logger
from kivy.properties import (AliasProperty, BooleanProperty, DictProperty,
								ListProperty, NumericProperty, ObjectProperty,
								OptionProperty, ReferenceListProperty,
								StringProperty, BoundedNumericProperty)
from kivy.graphics import InstructionGroup
from kivy.uix.boxlayout import BoxLayout
from kivy.uix.floatlayout import FloatLayout
from kivy.uix.layout import Layout
from kivy.uix.stencilview import StencilView
from kivy.uix.image import Image
from kivy.uix.widget import Widget

dbg = Logger.debug


def get_pos_hint_x(poshints, sizehintx):
	"""Return ``poshints['x']`` if available, or its computed equivalent
    otherwise.

    """
	if 'x' in poshints:
		return poshints['x']
	elif sizehintx is not None:
		if 'center_x' in poshints:
			return (poshints['center_x'] - sizehintx / 2)
		elif 'right' in poshints:
			return (poshints['right'] - sizehintx)


def get_pos_hint_y(poshints, sizehinty):
	"""Return ``poshints['y']`` if available, or its computed equivalent
    otherwise.

    """
	if 'y' in poshints:
		return poshints['y']
	elif sizehinty is not None:
		if 'center_y' in poshints:
			return (poshints['center_y'] - sizehinty / 2)
		elif 'top' in poshints:
			return (poshints['top'] - sizehinty)


def get_pos_hint(poshints, sizehintx, sizehinty):
	"""Return a tuple of ``(pos_hint_x, pos_hint_y)`` even if neither of
    those keys are present in the provided ``poshints`` -- they can be
    computed using the available keys together with ``size_hint_x``
    and ``size_hint_y``.

    """
	return (get_pos_hint_x(poshints,
							sizehintx), get_pos_hint_y(poshints, sizehinty))


class ColorTextureBox(Widget):
	"""A box, with a background of one solid color, an outline of another
    color, and possibly a texture covering the background.

    """
	color = ListProperty([1, 1, 1, 1])
	outline_color = ListProperty([0, 0, 0, 0])
	texture = ObjectProperty(None, allownone=True)


class Card(FloatLayout):
	"""A trading card, similar to the kind you use to play games like
    _Magic: the Gathering_.

    Its appearance is determined by several properties, the most
    important being:

    * ``headline_text``, a string to be shown at the top of the card;
      may be styled with eg. ``headline_font_name`` or
      ``headline_color``

    * ``art_source``, the path to an image to be displayed below the
      headline; may be hidden by setting ``show_art`` to ``False``

    * ``midline_text``, similar to ``headline_text`` but appearing
      below the art

    * ``text``, shown in a box the same size as the art. Styleable
      like ``headline_text`` and you can customize the box with
      eg. ``foreground_color`` and ``foreground_source``

    * ``footer_text``, like ``headline_text`` but at the bottom

    :class:`Card` is particularly useful when put in a
    :class:`DeckLayout`, allowing the user to drag cards in between
    any number of piles, into particular positions within a particular
    pile, and so forth.

    """
	dragging = BooleanProperty(False)
	deck = NumericProperty()
	idx = NumericProperty()
	ud = DictProperty({})

	collide_x = NumericProperty()
	collide_y = NumericProperty()
	collide_pos = ReferenceListProperty(collide_x, collide_y)

	foreground = ObjectProperty()
	foreground_source = StringProperty('')
	foreground_color = ListProperty([1, 1, 1, 1])
	foreground_image = ObjectProperty(None, allownone=True)
	foreground_texture = ObjectProperty(None, allownone=True)

	background_source = StringProperty('')
	background_color = ListProperty([.7, .7, .7, 1])
	background_image = ObjectProperty(None, allownone=True)
	background_texture = ObjectProperty(None, allownone=True)

	outline_color = ListProperty([0, 0, 0, 1])
	content_outline_color = ListProperty([0, 0, 0, 0])
	foreground_outline_color = ListProperty([0, 0, 0, 1])
	art_outline_color = ListProperty([0, 0, 0, 0])

	art = ObjectProperty()
	art_source = StringProperty('')
	art_color = ListProperty([1, 1, 1, 1])
	art_image = ObjectProperty(None, allownone=True)
	art_texture = ObjectProperty(None, allownone=True)
	show_art = BooleanProperty(True)

	headline = ObjectProperty()
	headline_text = StringProperty('Headline')
	headline_markup = BooleanProperty(True)
	headline_font_name = StringProperty('Roboto-Regular')
	headline_font_size = NumericProperty(18)
	headline_color = ListProperty([0, 0, 0, 1])

	midline = ObjectProperty()
	midline_text = StringProperty('')
	midline_markup = BooleanProperty(True)
	midline_font_name = StringProperty('Roboto-Regular')
	midline_font_size = NumericProperty(14)
	midline_color = ListProperty([0, 0, 0, 1])

	footer = ObjectProperty()
	footer_text = StringProperty('')
	footer_markup = BooleanProperty(True)
	footer_font_name = StringProperty('Roboto-Regular')
	footer_font_size = NumericProperty(10)
	footer_color = ListProperty([0, 0, 0, 1])

	text = StringProperty('')
	text_color = ListProperty([0, 0, 0, 1])
	markup = BooleanProperty(True)
	font_name = StringProperty('Roboto-Regular')
	font_size = NumericProperty(12)

	def on_background_source(self, *args):
		"""When I get a new ``background_source``, load it as an
        :class:`Image` and store that in ``background_image``.

        """
		if self.background_source:
			self.background_image = Image(source=self.background_source)

	def on_background_image(self, *args):
		"""When I get a new ``background_image``, store its texture in
        ``background_texture``.

        """
		if self.background_image is not None:
			self.background_texture = self.background_image.texture

	def on_foreground_source(self, *args):
		"""When I get a new ``foreground_source``, load it as an
        :class:`Image` and store that in ``foreground_image``.

        """
		if self.foreground_source:
			self.foreground_image = Image(source=self.foreground_source)

	def on_foreground_image(self, *args):
		"""When I get a new ``foreground_image``, store its texture in my
        ``foreground_texture``.

        """
		if self.foreground_image is not None:
			self.foreground_texture = self.foreground_image.texture

	def on_art_source(self, *args):
		"""When I get a new ``art_source``, load it as an :class:`Image` and
        store that in ``art_image``.

        """
		if self.art_source:
			self.art_image = Image(source=self.art_source)

	def on_art_image(self, *args):
		"""When I get a new ``art_image``, store its texture in
        ``art_texture``.

        """
		if self.art_image is not None:
			self.art_texture = self.art_image.texture

	def on_touch_down(self, touch):
		"""If I'm the first card to collide this touch, grab it, store my
        metadata in its userdict, and store the relative coords upon
        me where the collision happened.

        """
		if not self.collide_point(*touch.pos):
			return
		if 'card' in touch.ud:
			return
		touch.grab(self)
		self.dragging = True
		touch.ud['card'] = self
		touch.ud['idx'] = self.idx
		touch.ud['deck'] = self.deck
		touch.ud['layout'] = self.parent
		self.collide_x = touch.x - self.x
		self.collide_y = touch.y - self.y

	def on_touch_move(self, touch):
		"""If I'm being dragged, move so as to be always positioned the same
        relative to the touch.

        """
		if not self.dragging:
			touch.ungrab(self)
			return
		self.pos = (touch.x - self.collide_x, touch.y - self.collide_y)

	def on_touch_up(self, touch):
		"""Stop dragging if needed."""
		if not self.dragging:
			return
		touch.ungrab(self)
		self.dragging = False

	def copy(self):
		"""Return a new :class:`Card` just like me."""
		d = {}
		for att in ('deck', 'idx', 'ud', 'foreground_source',
					'foreground_color', 'foreground_image',
					'foreground_texture', 'background_source',
					'background_color', 'background_image',
					'background_texture', 'outline_color',
					'content_outline_color', 'foreground_outline_color',
					'art_outline_color', 'art_source', 'art_color',
					'art_image', 'art_texture', 'show_art', 'headline_text',
					'headline_markup', 'headline_font_name',
					'headline_font_size', 'headline_color', 'midline_text',
					'midline_markup', 'midline_font_name', 'midline_font_size',
					'midline_color', 'footer_text', 'footer_markup',
					'footer_font_name', 'footer_font_size', 'footer_color',
					'text', 'text_color', 'markup', 'font_name', 'font_size'):
			v = getattr(self, att)
			if v is not None:
				d[att] = v
		return Card(**d)


class Foundation(ColorTextureBox):
	"""An empty outline to indicate where a deck is when there are no
    cards in it.

    """
	color = ListProperty([])
	"""Color of the outline"""
	deck = NumericProperty(0)
	"""Index of the deck in the parent :class:`DeckLayout`"""

	def upd_pos(self, *args):
		"""Ask the foundation where I should be, based on what deck I'm
        for.

        """
		self.pos = self.parent._get_foundation_pos(self.deck)

	def upd_size(self, *args):
		"""I'm the same size as any given card in my :class:`DeckLayout`."""
		self.size = (self.parent.card_size_hint_x * self.parent.width,
						self.parent.card_size_hint_y * self.parent.height)


class DeckBuilderLayout(Layout):
	"""Sizes and positions :class:`Card` objects based on their order
    within ``decks``, a list of lists where each sublist is a deck of
    cards.

    """
	direction = OptionProperty('ascending',
								options=['ascending', 'descending'])
	"""Should the beginning card of each deck appear on the bottom
    ('ascending'), or the top ('descending')?

    """
	card_size_hint_x = BoundedNumericProperty(1, min=0, max=1)
	"""Each card's width, expressed as a proportion of my width."""
	card_size_hint_y = BoundedNumericProperty(1, min=0, max=1)
	"""Each card's height, expressed as a proportion of my height."""
	card_size_hint = ReferenceListProperty(card_size_hint_x, card_size_hint_y)
	"""Size hint of cards, relative to my size."""
	starting_pos_hint = DictProperty({'x': 0, 'y': 0})
	"""Pos hint at which to place the initial card of the initial deck."""
	card_x_hint_step = NumericProperty(0)
	"""Each time I put another card on a deck, I'll move it this much of
    my width to the right of the previous card.

    """
	card_y_hint_step = NumericProperty(-1)
	"""Each time I put another card on a deck, I'll move it this much of
    my height above the previous card.

    """
	card_hint_step = ReferenceListProperty(card_x_hint_step, card_y_hint_step)
	"""An offset, expressed in proportion to my size, applied to each
    successive card in a given deck.

    """
	deck_x_hint_step = NumericProperty(1)
	"""When I start a new deck, it will be this far to the right of the
    previous deck, expressed as a proportion of my width.

    """
	deck_y_hint_step = NumericProperty(0)
	"""When I start a new deck, it will be this far above the previous
    deck, expressed as a proportion of my height.

    """
	deck_hint_step = ReferenceListProperty(deck_x_hint_step, deck_y_hint_step)
	"""Offset of each deck with respect to the previous, as a proportion
    of my size.

    """
	decks = ListProperty([[]])  # list of lists of cards
	"""Put a list of lists of :class:`Card` objects here and I'll position
    them appropriately. Please don't use ``add_widget``.

    """
	deck_x_hint_offsets = ListProperty([])
	"""An additional proportional x-offset for each deck, defaulting to 0."""
	deck_y_hint_offsets = ListProperty([])
	"""An additional proportional y-offset for each deck, defaulting to 0."""
	foundation_color = ListProperty([1, 1, 1, 1])
	"""Color to use for the outline showing where a deck is when it's
    empty.

    """
	insertion_deck = BoundedNumericProperty(None, min=0, allownone=True)
	"""Index of the deck that a card is being dragged into."""
	insertion_card = BoundedNumericProperty(None, min=0, allownone=True)
	"""Index within the current deck that a card is being dragged into."""
	_foundations = ListProperty([])
	"""Private. A list of :class:`Foundation` widgets, one per deck."""

	def __init__(self, **kwargs):
		"""Bind most of my custom properties to ``_trigger_layout``."""
		super().__init__(**kwargs)
		self.bind(card_size_hint=self._trigger_layout,
					starting_pos_hint=self._trigger_layout,
					card_hint_step=self._trigger_layout,
					deck_hint_step=self._trigger_layout,
					decks=self._trigger_layout,
					deck_x_hint_offsets=self._trigger_layout,
					deck_y_hint_offsets=self._trigger_layout,
					insertion_deck=self._trigger_layout,
					insertion_card=self._trigger_layout)

	def scroll_deck_x(self, decknum, scroll_x):
		"""Move a deck left or right."""
		if decknum >= len(self.decks):
			raise IndexError("I have no deck at {}".format(decknum))
		if decknum >= len(self.deck_x_hint_offsets):
			self.deck_x_hint_offsets = list(self.deck_x_hint_offsets) + [0] * (
				decknum - len(self.deck_x_hint_offsets) + 1)
		self.deck_x_hint_offsets[decknum] += scroll_x
		self._trigger_layout()

	def scroll_deck_y(self, decknum, scroll_y):
		"""Move a deck up or down."""
		if decknum >= len(self.decks):
			raise IndexError("I have no deck at {}".format(decknum))
		if decknum >= len(self.deck_y_hint_offsets):
			self.deck_y_hint_offsets = list(self.deck_y_hint_offsets) + [0] * (
				decknum - len(self.deck_y_hint_offsets) + 1)
		self.deck_y_hint_offsets[decknum] += scroll_y
		self._trigger_layout()

	def scroll_deck(self, decknum, scroll_x, scroll_y):
		"""Move a deck."""
		self.scroll_deck_x(decknum, scroll_x)
		self.scroll_deck_y(decknum, scroll_y)

	def _get_foundation_pos(self, i):
		"""Private. Get the absolute coordinates to use for a deck's
        foundation, based on the ``starting_pos_hint``, the
        ``deck_hint_step``, ``deck_x_hint_offsets``, and
        ``deck_y_hint_offsets``.

        """
		(phx, phy) = get_pos_hint(self.starting_pos_hint, *self.card_size_hint)
		phx += self.deck_x_hint_step * i + self.deck_x_hint_offsets[i]
		phy += self.deck_y_hint_step * i + self.deck_y_hint_offsets[i]
		x = phx * self.width + self.x
		y = phy * self.height + self.y
		return (x, y)

	def _get_foundation(self, i):
		"""Return a :class:`Foundation` for some deck, creating it if
        needed.

        """
		if i >= len(self._foundations) or self._foundations[i] is None:
			oldfound = list(self._foundations)
			extend = i - len(oldfound) + 1
			if extend > 0:
				oldfound += [None] * extend
			width = self.card_size_hint_x * self.width
			height = self.card_size_hint_y * self.height
			found = Foundation(pos=self._get_foundation_pos(i),
								size=(width, height),
								deck=i)
			self.bind(pos=found.upd_pos,
						card_size_hint=found.upd_pos,
						deck_hint_step=found.upd_pos,
						size=found.upd_pos,
						deck_x_hint_offsets=found.upd_pos,
						deck_y_hint_offsets=found.upd_pos)
			self.bind(size=found.upd_size, card_size_hint=found.upd_size)
			oldfound[i] = found
			self._foundations = oldfound
		return self._foundations[i]

	def on_decks(self, *args):
		"""Inform the cards of their deck and their index within the deck;
        extend the ``_hint_offsets`` properties as needed; and trigger
        a layout.

        """
		if None in (self.canvas, self.decks, self.deck_x_hint_offsets,
					self.deck_y_hint_offsets):
			Clock.schedule_once(self.on_decks, 0)
			return
		self.clear_widgets()
		decknum = 0
		for deck in self.decks:
			cardnum = 0
			for card in deck:
				if not isinstance(card, Card):
					raise TypeError("You must only put Card in decks")
				if card not in self.children:
					self.add_widget(card)
				if card.deck != decknum:
					card.deck = decknum
				if card.idx != cardnum:
					card.idx = cardnum
				cardnum += 1
			decknum += 1
		if len(self.deck_x_hint_offsets) < len(self.decks):
			self.deck_x_hint_offsets = list(self.deck_x_hint_offsets) + [0] * (
				len(self.decks) - len(self.deck_x_hint_offsets))
		if len(self.deck_y_hint_offsets) < len(self.decks):
			self.deck_y_hint_offsets = list(self.deck_y_hint_offsets) + [0] * (
				len(self.decks) - len(self.deck_y_hint_offsets))
		self._trigger_layout()

	def point_before_card(self, card, x, y):
		"""Return whether ``(x, y)`` is somewhere before ``card``, given how I
        know cards to be arranged.

        If the cards are being stacked down and to the right, that
        means I'm testing whether ``(x, y)`` is above or to the left
        of the card.

        """

		def ycmp():
			if self.card_y_hint_step == 0:
				return False
			elif self.card_y_hint_step > 0:
				# stacking upward
				return y < card.y
			else:
				# stacking downward
				return y > card.top

		if self.card_x_hint_step > 0:
			# stacking to the right
			if x < card.x:
				return True
			return ycmp()
		elif self.card_x_hint_step == 0:
			return ycmp()
		else:
			# stacking to the left
			if x > card.right:
				return True
			return ycmp()

	def point_after_card(self, card, x, y):
		"""Return whether ``(x, y)`` is somewhere after ``card``, given how I
        know cards to be arranged.

        If the cards are being stacked down and to the right, that
        means I'm testing whether ``(x, y)`` is below or to the left
        of ``card``.

        """

		def ycmp():
			if self.card_y_hint_step == 0:
				return False
			elif self.card_y_hint_step > 0:
				# stacking upward
				return y > card.top
			else:
				# stacking downward
				return y < card.y

		if self.card_x_hint_step > 0:
			# stacking to the right
			if x > card.right:
				return True
			return ycmp()
		elif self.card_x_hint_step == 0:
			return ycmp()
		else:
			# stacking to the left
			if x < card.x:
				return True
			return ycmp()

	def on_touch_move(self, touch):
		"""If a card is being dragged, move other cards out of the way to show
        where the dragged card will go if you drop it.

        """
		if ('card' not in touch.ud or 'layout' not in touch.ud
			or touch.ud['layout'] != self):
			return
		if (touch.ud['layout'] == self
			and not hasattr(touch.ud['card'], '_topdecked')):
			touch.ud['card']._topdecked = InstructionGroup()
			touch.ud['card']._topdecked.add(touch.ud['card'].canvas)
			self.canvas.after.add(touch.ud['card']._topdecked)
		for i, deck in enumerate(self.decks):
			cards = [card for card in deck if not card.dragging]
			maxidx = max(card.idx for card in cards) if cards else 0
			if self.direction == 'descending':
				cards.reverse()
			cards_collided = [
				card for card in cards if card.collide_point(*touch.pos)
			]
			if cards_collided:
				collided = cards_collided.pop()
				for card in cards_collided:
					if card.idx > collided.idx:
						collided = card
				if collided.deck == touch.ud['deck']:
					self.insertion_card = (
						1 if collided.idx == 0 else maxidx +
						1 if collided.idx == maxidx else collided.idx +
						1 if collided.idx > touch.ud['idx'] else collided.idx)
				else:
					dropdeck = self.decks[collided.deck]
					maxidx = max(card.idx for card in dropdeck)
					self.insertion_card = (
						1 if collided.idx == 0 else maxidx +
						1 if collided.idx == maxidx else collided.idx + 1)
				if self.insertion_deck != collided.deck:
					self.insertion_deck = collided.deck
				return
			else:
				if self.insertion_deck == i:
					if self.insertion_card in (0, len(deck)):
						pass
					elif self.point_before_card(cards[0], *touch.pos):
						self.insertion_card = 0
					elif self.point_after_card(cards[-1], *touch.pos):
						self.insertion_card = cards[-1].idx
				else:
					for j, found in enumerate(self._foundations):
						if (found is not None
							and found.collide_point(*touch.pos)):
							self.insertion_deck = j
							self.insertion_card = 0
							return

	def on_touch_up(self, touch):
		"""If a card is being dragged, put it in the place it was just dropped
        and trigger a layout.

        """
		if ('card' not in touch.ud or 'layout' not in touch.ud
			or touch.ud['layout'] != self):
			return
		if hasattr(touch.ud['card'], '_topdecked'):
			self.canvas.after.remove(touch.ud['card']._topdecked)
			del touch.ud['card']._topdecked
		if None not in (self.insertion_deck, self.insertion_card):
			# need to sync to adapter.data??
			card = touch.ud['card']
			del card.parent.decks[card.deck][card.idx]
			for i in range(0, len(card.parent.decks[card.deck])):
				card.parent.decks[card.deck][i].idx = i
			deck = self.decks[self.insertion_deck]
			if self.insertion_card >= len(deck):
				deck.append(card)
			else:
				deck.insert(self.insertion_card, card)
			card.deck = self.insertion_deck
			card.idx = self.insertion_card
			self.decks[self.insertion_deck] = deck
			self.insertion_deck = self.insertion_card = None
		self._trigger_layout()

	def on_insertion_card(self, *args):
		"""Trigger a layout"""
		if self.insertion_card is not None:
			self._trigger_layout()

	def do_layout(self, *args):
		"""Layout each of my decks"""
		if self.size == [1, 1]:
			return
		for i in range(0, len(self.decks)):
			self.layout_deck(i)

	def layout_deck(self, i):
		"""Stack the cards, starting at my deck's foundation, and proceeding
        by ``card_pos_hint``

        """

		def get_dragidx(cards):
			j = 0
			for card in cards:
				if card.dragging:
					return j
				j += 1

		# Put a None in the card list in place of the card you're
		# hovering over, if you're dragging another card. This will
		# result in an empty space where the card will go if you drop
		# it now.
		cards = list(self.decks[i])
		dragidx = get_dragidx(cards)
		if dragidx is not None:
			del cards[dragidx]
		if self.insertion_deck == i and self.insertion_card is not None:
			insdx = self.insertion_card
			if dragidx is not None and insdx > dragidx:
				insdx -= 1
			cards.insert(insdx, None)
		if self.direction == 'descending':
			cards.reverse()
		# Work out the initial pos_hint for this deck
		(phx, phy) = get_pos_hint(self.starting_pos_hint, *self.card_size_hint)
		phx += self.deck_x_hint_step * i + self.deck_x_hint_offsets[i]
		phy += self.deck_y_hint_step * i + self.deck_y_hint_offsets[i]
		(w, h) = self.size
		(x, y) = self.pos
		# start assigning pos and size to cards
		found = self._get_foundation(i)
		if found in self.children:
			self.remove_widget(found)
		self.add_widget(found)
		for card in cards:
			if card is not None:
				if card in self.children:
					self.remove_widget(card)
				(shw, shh) = self.card_size_hint
				card.pos = (x + phx * w, y + phy * h)
				card.size = (w * shw, h * shh)
				self.add_widget(card)
			phx += self.card_x_hint_step
			phy += self.card_y_hint_step


class DeckBuilderView(DeckBuilderLayout, StencilView):
	"""Just a :class:`DeckBuilderLayout` mixed with
    :class:`StencilView`.

    """
	pass


class ScrollBarBar(ColorTextureBox):
	"""Tiny tweak to :class:`ColorTextureBox` to make it work within
    :class:`DeckBuilderScrollBar`

    """

	def on_touch_down(self, touch):
		"""Tell my parent if I've been touched"""
		if self.parent is None:
			return
		if self.collide_point(*touch.pos):
			self.parent.bar_touched(self, touch)


class DeckBuilderScrollBar(FloatLayout):
	"""A widget that looks a lot like one of the scrollbars on the sides
    of eg. :class:`kivy.uix.ScrollView`, which moves a single deck
    within a :class:`DeckBuilderLayout`.

    """
	orientation = OptionProperty('vertical',
									options=['horizontal', 'vertical'])
	"""Which way to scroll? Options are 'horizontal' and 'vertical'."""
	deckbuilder = ObjectProperty()
	"""The :class:`DeckBuilderLayout` of the deck to scroll."""
	deckidx = NumericProperty(0)
	"""The index of the deck to scroll, within its
    :class:`DeckBuilderLayout`'s ``decks`` property.

    """
	scrolling = BooleanProperty(False)
	"""Has the user grabbed me?"""
	scroll_min = NumericProperty(-1)
	"""How far left (if horizontal) or down (if vertical) I can move my
    deck, expressed as a proportion of the
    :class:`DeckBuilderLayout`'s width or height, respectively.

    """
	scroll_max = NumericProperty(1)
	"""How far right (if horizontal) or up (if vertical) I can move my
    deck, expressed as a proportion of the
    :class:`DeckBuilderLayout`'s width or height, respectively.

    """

	scroll_hint = AliasProperty(
		lambda self: abs(self.scroll_max - self.scroll_min),
		lambda self, v: None,
		bind=('scroll_min', 'scroll_max'))
	"""The distance between ``scroll_max`` and ``scroll_min``."""
	_scroll = NumericProperty(0)
	"""Private. The current adjustment to the deck's ``pos_hint_x`` or
    ``pos_hint_y``.

    """

	def _get_scroll(self):
		zero = self._scroll - self.scroll_min
		return zero / self.scroll_hint

	def _set_scroll(self, v):
		if v < 0:
			v = 0
		if v > 1:
			v = 1
		normal = v * self.scroll_hint
		self._scroll = self.scroll_min + normal

	scroll = AliasProperty(_get_scroll,
							_set_scroll,
							bind=('_scroll', 'scroll_min', 'scroll_max'))
	"""A number between 0 and 1 representing how far beyond ``scroll_min``
    toward ``scroll_max`` I am presently scrolled.

    """

	def _get_vbar(self):
		if self.deckbuilder is None:
			return (0, 1)
		vh = self.deckbuilder.height * (self.scroll_hint + 1)
		h = self.height
		if vh < h or vh == 0:
			return (0, 1)
		ph = max(0.01, h / vh)
		sy = min(1.0, max(0.0, self.scroll))
		py = (1 - ph) * sy
		return (py, ph)

	vbar = AliasProperty(_get_vbar,
							None,
							bind=('_scroll', 'scroll_min', 'scroll_max'))
	"""A tuple of ``(y, height)`` for my scroll bar, if it's vertical."""

	def _get_hbar(self):
		if self.deckbuilder is None:
			return (0, 1)
		vw = self.deckbuilder.width * self.scroll_hint
		w = self.width
		if vw < w or vw == 0:
			return (0, 1)
		pw = max(0.01, w / vw)
		sx = min(1.0, max(0.0, self.scroll))
		px = (1 - pw) * sx
		return (px, pw)

	hbar = AliasProperty(_get_hbar,
							None,
							bind=('_scroll', 'scroll_min', 'scroll_max'))
	"""A tuple of ``(x, width)`` for my scroll bar, if it's horizontal."""
	bar_color = ListProperty([.7, .7, .7, .9])
	"""Color to use for the scroll bar when scrolling. RGBA format."""
	bar_inactive_color = ListProperty([.7, .7, .7, .2])
	"""Color to use for the scroll bar when not scrolling. RGBA format."""
	bar_texture = ObjectProperty(None, allownone=True)
	"""Texture for the scroll bar, normally ``None``."""

	def __init__(self, **kwargs):
		"""Arrange to be laid out whenever I'm scrolled or the range of my
        scrolling changes.

        """
		super().__init__(**kwargs)
		self.bind(_scroll=self._trigger_layout,
					scroll_min=self._trigger_layout,
					scroll_max=self._trigger_layout)

	def do_layout(self, *args):
		"""Put the bar where it's supposed to be, and size it in proportion to
        the size of the scrollable area.

        """
		if 'bar' not in self.ids:
			Clock.schedule_once(self.do_layout)
			return
		if self.orientation == 'horizontal':
			self.ids.bar.size_hint_x = self.hbar[1]
			self.ids.bar.pos_hint = {'x': self.hbar[0], 'y': 0}
		else:
			self.ids.bar.size_hint_y = self.vbar[1]
			self.ids.bar.pos_hint = {'x': 0, 'y': self.vbar[0]}
		super().do_layout(*args)

	def upd_scroll(self, *args):
		"""Update my own ``scroll`` property to where my deck is actually
        scrolled.

        """
		att = 'deck_{}_hint_offsets'.format('x' if self.orientation ==
											'horizontal' else 'y')
		self._scroll = getattr(self.deckbuilder, att)[self.deckidx]

	def on_deckbuilder(self, *args):
		"""Bind my deckbuilder to update my ``scroll``, and my ``scroll`` to
        update my deckbuilder.

        """
		if self.deckbuilder is None:
			return
		att = 'deck_{}_hint_offsets'.format('x' if self.orientation ==
											'horizontal' else 'y')
		offs = getattr(self.deckbuilder, att)
		if len(offs) <= self.deckidx:
			Clock.schedule_once(self.on_deckbuilder, 0)
			return
		self.bind(scroll=self.handle_scroll)
		self.deckbuilder.bind(**{att: self.upd_scroll})
		self.upd_scroll()
		self.deckbuilder._trigger_layout()

	def handle_scroll(self, *args):
		"""When my ``scroll`` changes, tell my deckbuilder how it's scrolled
        now.

        """
		if 'bar' not in self.ids:
			Clock.schedule_once(self.handle_scroll, 0)
			return
		att = 'deck_{}_hint_offsets'.format('x' if self.orientation ==
											'horizontal' else 'y')
		offs = list(getattr(self.deckbuilder, att))
		if len(offs) <= self.deckidx:
			Clock.schedule_once(self.on_scroll, 0)
			return
		offs[self.deckidx] = self._scroll
		setattr(self.deckbuilder, att, offs)
		self.deckbuilder._trigger_layout()

	def bar_touched(self, bar, touch):
		"""Start scrolling, and record where I started scrolling."""
		self.scrolling = True
		self._start_bar_pos_hint = get_pos_hint(bar.pos_hint, *bar.size_hint)
		self._start_touch_pos_hint = (touch.x / self.width,
										touch.y / self.height)
		self._start_bar_touch_hint = (self._start_touch_pos_hint[0] -
										self._start_bar_pos_hint[0],
										self._start_touch_pos_hint[1] -
										self._start_bar_pos_hint[1])
		touch.grab(self)

	def on_touch_move(self, touch):
		"""Move the scrollbar to the touch, and update my ``scroll``
        accordingly.

        """
		if not self.scrolling or 'bar' not in self.ids:
			touch.ungrab(self)
			return
		touch.push()
		touch.apply_transform_2d(self.parent.to_local)
		touch.apply_transform_2d(self.to_local)
		if self.orientation == 'horizontal':
			hint_right_of_bar = (touch.x - self.ids.bar.x) / self.width
			hint_correction = hint_right_of_bar - self._start_bar_touch_hint[0]
			self.scroll += hint_correction
		else:  # self.orientation == 'vertical'
			hint_above_bar = (touch.y - self.ids.bar.y) / self.height
			hint_correction = hint_above_bar - self._start_bar_touch_hint[1]
			self.scroll += hint_correction
		touch.pop()

	def on_touch_up(self, touch):
		"""Stop scrolling."""
		self.scrolling = False


kv = """
<ColorTextureBox>:
    canvas:
        Color:
            rgba: root.color
        Rectangle:
            texture: root.texture
            pos: root.pos
            size: root.size
        Color:
            rgba: root.outline_color
        Line:
            points: [self.x, self.y, self.right, self.y, self.right, self.top, self.x, self.top, self.x, self.y]
        Color:
            rgba: [1, 1, 1, 1]
<Foundation>:
    color: [0, 0, 0, 0]
    outline_color: [1, 1, 1, 1]
<Card>:
    headline: headline
    midline: midline
    footer: footer
    art: art
    foreground: foreground
    canvas:
        Color:
            rgba: root.background_color
        Rectangle:
            texture: root.background_texture
            pos: root.pos
            size: root.size
        Color:
            rgba: root.outline_color
        Line:
            points: [self.x, self.y, self.right, self.y, self.right, self.top, self.x, self.top, self.x, self.y]
        Color:
            rgba: [1, 1, 1, 1]
    BoxLayout:
        size_hint: 0.9, 0.9
        pos_hint: {'x': 0.05, 'y': 0.05}
        orientation: 'vertical'
        canvas:
            Color:
                rgba: root.content_outline_color
            Line:
                points: [self.x, self.y, self.right, self.y, self.right, self.top, self.x, self.top, self.x, self.y]
            Color:
                rgba: [1, 1, 1, 1]
        Label:
            id: headline
            text: root.headline_text
            markup: root.headline_markup
            font_name: root.headline_font_name
            font_size: root.headline_font_size
            color: root.headline_color
            size_hint: (None, None)
            size: self.texture_size
        ColorTextureBox:
            id: art
            color: root.art_color
            texture: root.art_texture
            outline_color: root.art_outline_color if root.show_art else [0, 0, 0, 0]
            size_hint: (1, 1) if root.show_art else (None, None)
            size: (0, 0)
        Label:
            id: midline
            text: root.midline_text
            markup: root.midline_markup
            font_name: root.midline_font_name
            font_size: root.midline_font_size
            color: root.midline_color
            size_hint: (None, None)
            size: self.texture_size
        ColorTextureBox:
            id: foreground
            color: root.foreground_color
            outline_color: root.foreground_outline_color
            texture: root.foreground_texture
            Label:
                text: root.text
                color: root.text_color
                markup: root.markup
                font_name: root.font_name
                font_size: root.font_size
                text_size: foreground.size
                size_hint: (None, None)
                size: self.texture_size
                pos: foreground.pos
                valign: 'top'
        Label:
            id: footer
            text: root.footer_text
            markup: root.footer_markup
            font_name: root.footer_font_name
            font_size: root.footer_font_size
            color: root.footer_color
            size_hint: (None, None)
            size: self.texture_size
<DeckBuilderScrollBar>:
    ScrollBarBar:
        id: bar
        color: root.bar_color if root.scrolling else root.bar_inactive_color
        texture: root.bar_texture
"""
Builder.load_string(kv)

if __name__ == '__main__':
	deck0 = [
		Card(background_color=[0, 1, 0, 1],
				headline_text='Card {}'.format(i),
				art_color=[1, 0, 0, 1],
				midline_text='0deck',
				foreground_color=[0, 0, 1, 1],
				text='The quick brown fox jumps over the lazy dog',
				text_color=[1, 1, 1, 1],
				footer_text=str(i)) for i in range(0, 9)
	]
	deck1 = [
		Card(background_color=[0, 0, 1, 1],
				headline_text='Card {}'.format(i),
				art_color=[0, 1, 0, 1],
				show_art=False,
				midline_text='1deck',
				foreground_color=[1, 0, 0, 1],
				text='Have a steak at the porter house bar',
				text_color=[1, 1, 0, 1],
				footer_text=str(i)) for i in range(0, 9)
	]
	from kivy.base import runTouchApp
	from kivy.core.window import Window
	from kivy.modules import inspector

	builder = DeckBuilderLayout(card_size_hint=(0.15, 0.3),
								pos_hint={
									'x': 0,
									'y': 0
								},
								starting_pos_hint={
									'x': 0.1,
									'top': 0.9
								},
								card_hint_step=(0.05, -0.1),
								deck_hint_step=(0.4, 0),
								decks=[deck0, deck1],
								deck_y_hint_offsets=[0, 1])
	layout = BoxLayout(orientation='horizontal')
	left_bar = DeckBuilderScrollBar(deckbuilder=builder,
									orientation='vertical',
									size_hint_x=0.1,
									deckidx=0)
	right_bar = DeckBuilderScrollBar(deckbuilder=builder,
										orientation='vertical',
										size_hint_x=0.1,
										deckidx=1)
	layout.add_widget(left_bar)
	layout.add_widget(builder)
	layout.add_widget(right_bar)
	inspector.create_inspector(Window, layout)
	runTouchApp(layout)



# This file is part of ELiDE, frontend to LiSE, a framework for life simulation games.
# Copyright (c) Zachary Spector, public@zacharyspector.com
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, version 3.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""Object to configure, start, and stop ELiDE."""
import sys
import os
import json
from functools import partial

from LiSE.allegedb import OutOfTimelineError

if 'KIVY_NO_ARGS' not in os.environ:
	os.environ['KIVY_NO_ARGS'] = '1'

from kivy.logger import Logger
from kivy.app import App
from kivy.clock import Clock
from kivy.resources import resource_add_path

from kivy.uix.screenmanager import ScreenManager, NoTransition

from kivy.properties import (AliasProperty, BooleanProperty, ObjectProperty,
								NumericProperty, StringProperty)
import LiSE
from LiSE.proxy import EngineProcessManager, CharStatProxy
import ELiDE
import ELiDE.dialog
import ELiDE.screen
import ELiDE.stores
import ELiDE.statcfg
import ELiDE.spritebuilder
import ELiDE.rulesview
import ELiDE.charsview
import ELiDE.timestream
from ELiDE.graph.board import GraphBoard
from ELiDE.graph.arrow import GraphArrowWidget
from ELiDE.graph.spot import GraphSpot
from ELiDE.graph.pawn import Pawn
from ELiDE.grid.board import GridBoard
from .util import trigger

resource_add_path(ELiDE.__path__[0] + "/assets")
resource_add_path(ELiDE.__path__[0] + "/assets/rltiles")


class ELiDEApp(App):
	"""Extensible LiSE Development Environment.

    """
	title = 'ELiDE'

	engine = None
	branch = StringProperty('trunk')
	turn = NumericProperty(0)
	tick = NumericProperty(0)
	character = ObjectProperty()
	selection = ObjectProperty(None, allownone=True)
	selected_proxy = ObjectProperty()
	selected_proxy_name = StringProperty('')
	statcfg = ObjectProperty()
	edit_locked = BooleanProperty(False)

	def on_selection(self, *args):
		Logger.debug("App: {} selected".format(self.selection))

	def on_selected_proxy(self, *args):
		if hasattr(self.selected_proxy, 'name'):
			self.selected_proxy_name = str(self.selected_proxy.name)
			return
		selected_proxy = self.selected_proxy
		assert hasattr(selected_proxy, 'origin'), '{} has no origin'.format(
			type(selected_proxy))
		assert hasattr(selected_proxy,
						'destination'), '{} has no destination'.format(
							type(selected_proxy))
		origin = selected_proxy.origin
		destination = selected_proxy.destination
		reciprocal = selected_proxy.reciprocal
		if selected_proxy.get(
			'is_mirror', False) or (reciprocal
									and reciprocal.get('is_mirror', False)):
			link = '<>'
		else:
			link = '->'
		self.selected_proxy_name = str(origin.name) + link + str(
			destination.name)

	def _get_character_name(self, *args):
		if self.character is None:
			return
		return self.character.name

	def _set_character_name(self, name):
		if self.character.name != name:
			self.character = self.engine.character[name]

	character_name = AliasProperty(_get_character_name,
									_set_character_name,
									bind=('character', ))

	def _pull_time(self, *args):
		if not self.engine:
			Clock.schedule_once(self._pull_time, 0)
			return
		branch, turn, tick = self.engine._btt()
		self.branch = branch
		self.turn = turn
		self.tick = tick

	pull_time = trigger(_pull_time)

	def time_travel(self, branch, turn, tick=None):
		try:
			self.engine.time_travel(branch,
									turn,
									tick,
									chars='all',
									cb=self._update_from_time_travel)
		except OutOfTimelineError as ex:
			(self.branch, self.turn, self.tick) = (ex.branch_from,
													ex.turn_from, ex.tick_from)

	def _update_from_time_travel(self, command, branch, turn, tick, result,
									**kwargs):
		(self.branch, self.turn, self.tick) = (branch, turn, tick)
		self.mainscreen._update_from_time_travel(command, branch, turn, tick,
													result, **kwargs)

	def set_tick(self, t):
		"""Set my tick to the given value, cast to an integer."""
		self.tick = int(t)

	def set_turn(self, t):
		"""Set the turn to the given value, cast to an integer"""
		self.turn = int(t)

	def select_character(self, char):
		"""Change my ``character`` to the selected character object if they
        aren't the same.

        """
		if char == self.character:
			return
		self.character = char

	def build_config(self, config):
		"""Set config defaults"""
		for sec in 'LiSE', 'ELiDE':
			config.adddefaultsection(sec)
		config.setdefaults(
			'LiSE', {
				'world': 'sqlite:///world.db',
				'language': 'eng',
				'logfile': '',
				'loglevel': 'info'
			})
		config.setdefaults(
			'ELiDE', {
				'debugger':
				'no',
				'inspector':
				'no',
				'user_kv':
				'yes',
				'play_speed':
				'1',
				'thing_graphics':
				json.dumps([('RLTiles: Body', 'base.atlas'),
							('RLTiles: Basic clothes', 'body.atlas'),
							('RLTiles: Armwear', 'arm.atlas'),
							('RLTiles: Legwear', 'leg.atlas'),
							('RLTiles: Right hand', 'hand1.atlas'),
							('RLTiles: Left hand', 'hand2.atlas'),
							('RLTiles: Boots', 'boot.atlas'),
							('RLTiles: Hair', 'hair.atlas'),
							('RLTiles: Beard', 'beard.atlas'),
							('RLTiles: Headwear', 'head.atlas')]),
				'place_graphics':
				json.dumps([('RLTiles: Dungeon', 'dungeon.atlas'),
							('RLTiles: Floor', 'floor.atlas')])
			})
		config.write()

	def build(self):
		self.icon = 'icon_24px.png'
		config = self.config
		Logger.debug("ELiDEApp: starting with world {}, path {}".format(
			config['LiSE']['world'], LiSE.__path__[-1]))

		if config['ELiDE']['debugger'] == 'yes':
			import pdb
			pdb.set_trace()

		self.manager = ScreenManager(transition=NoTransition())
		if config['ELiDE']['inspector'] == 'yes':
			from kivy.core.window import Window
			from kivy.modules import inspector
			inspector.create_inspector(Window, self.manager)

		self._add_screens()
		return self.manager

	def _pull_lang(self, *args, **kwargs):
		self.strings.language = kwargs['language']

	def _pull_chars(self, *args, **kwargs):
		self.chars.names = list(self.engine.character)

	def _pull_time_from_signal(self, *args, branch, turn, tick):
		self.branch, self.turn, self.tick = branch, turn, tick

	def start_subprocess(self, *args):
		"""Start the LiSE core and get a proxy to it

        Must be called before ``init_board``

        """
		if hasattr(self, '_started'):
			raise ChildProcessError("Subprocess already running")
		config = self.config
		enkw = {'logger': Logger}
		if config['LiSE'].get('logfile'):
			enkw['logfile'] = config['LiSE']['logfile']
		if config['LiSE'].get('loglevel'):
			enkw['loglevel'] = config['LiSE']['loglevel']
		self.procman = EngineProcessManager()
		self.engine = engine = self.procman.start(**enkw)
		self.pull_time()

		self.engine.time.connect(self._pull_time_from_signal, weak=False)
		self.engine.string.language.connect(self._pull_lang, weak=False)
		self.engine.character.connect(self._pull_chars, weak=False)

		self.strings.store = self.engine.string
		self._started = True
		return engine

	trigger_start_subprocess = trigger(start_subprocess)

	def init_board(self, *args):
		"""Get the board widgets initialized to display the game state

        Must be called after start_subprocess

        """
		if 'boardchar' not in self.engine.eternal:
			if 'physical' in self.engine.character:
				self.engine.eternal['boardchar'] = self.engine.character[
					'physical']
			else:
				chara = self.engine.eternal[
					'boardchar'] = self.engine.new_character('physical')
		self.chars.names = list(self.engine.character)
		self.mainscreen.graphboards = {
			name: GraphBoard(character=char)
			for name, char in self.engine.character.items()
		}
		self.mainscreen.gridboards = {
			name: GridBoard(character=char)
			for name, char in self.engine.character.items()
		}
		self.select_character(self.engine.eternal['boardchar'])
		self.selected_proxy = self._get_selected_proxy()

	def _add_screens(self, *args):

		def toggler(screenname):

			def tog(*args):
				if self.manager.current == screenname:
					self.manager.current = 'main'
				else:
					self.manager.current = screenname

			return tog

		config = self.config

		self.mainmenu = ELiDE.menu.DirPicker(toggle=toggler('mainmenu'))

		self.pawncfg = ELiDE.spritebuilder.PawnConfigScreen(
			toggle=toggler('pawncfg'),
			data=json.loads(config['ELiDE']['thing_graphics']))

		self.spotcfg = ELiDE.spritebuilder.SpotConfigScreen(
			toggle=toggler('spotcfg'),
			data=json.loads(config['ELiDE']['place_graphics']))

		self.statcfg = ELiDE.statcfg.StatScreen(toggle=toggler('statcfg'))

		self.rules = ELiDE.rulesview.RulesScreen(toggle=toggler('rules'))

		self.charrules = ELiDE.rulesview.CharacterRulesScreen(
			character=self.character, toggle=toggler('charrules'))
		self.bind(character=self.charrules.setter('character'))

		self.chars = ELiDE.charsview.CharactersScreen(
			toggle=toggler('chars'), new_board=self.new_board)
		self.bind(character_name=self.chars.setter('character_name'))

		def chars_push_character_name(*args):
			self.unbind(character_name=self.chars.setter('character_name'))
			self.character_name = self.chars.character_name
			self.bind(character_name=self.chars.setter('character_name'))

		self.chars.push_character_name = chars_push_character_name

		self.strings = ELiDE.stores.StringsEdScreen(toggle=toggler('strings'))

		self.funcs = ELiDE.stores.FuncsEdScreen(name='funcs',
												toggle=toggler('funcs'))

		self.bind(selected_proxy=self.statcfg.setter('proxy'))

		self.timestream = ELiDE.timestream.TimestreamScreen(
			name='timestream', toggle=toggler('timestream'))

		self.mainscreen = ELiDE.screen.MainScreen(
			use_kv=config['ELiDE']['user_kv'] == 'yes',
			play_speed=int(config['ELiDE']['play_speed']))
		if self.mainscreen.statlist:
			self.statcfg.statlist = self.mainscreen.statlist
		self.mainscreen.bind(statlist=self.statcfg.setter('statlist'))
		self.bind(selection=self.refresh_selected_proxy,
					character=self.refresh_selected_proxy)
		for wid in (self.mainmenu, self.mainscreen, self.pawncfg, self.spotcfg,
					self.statcfg, self.rules, self.charrules, self.chars,
					self.strings, self.funcs, self.timestream):
			self.manager.add_widget(wid)
		if (os.environ['KIVY_NO_ARGS']
			or sys.argv[-2] == '-') and os.path.exists(
				sys.argv[-1]) and os.path.isdir(sys.argv[-1]):
			Clock.schedule_once(
				partial(self.mainmenu.open, os.path.abspath(sys.argv[-1])),
				0.001)

	def update_calendar(self, calendar, past_turns=1, future_turns=5):
		"""Fill in a calendar widget with actual simulation data"""
		startturn = self.turn - past_turns
		endturn = self.turn + future_turns
		stats = ['_config'] + [
			stat for stat in self.selected_proxy
			if not stat.startswith('_') and stat not in ('character', 'name')
		]
		if isinstance(self.selected_proxy, CharStatProxy):
			sched_entity = self.engine.character[self.selected_proxy.name]
		else:
			sched_entity = self.selected_proxy
		calendar.entity = sched_entity
		calendar.from_schedule(self.engine.handle('get_schedule',
													entity=sched_entity,
													stats=stats,
													beginning=startturn,
													end=endturn),
								start_turn=startturn)

	def _set_language(self, lang):
		self.engine.string.language = lang

	def _get_selected_proxy(self):
		if self.selection is None:
			return self.character.stat
		elif hasattr(self.selection, 'proxy'):
			return self.selection.proxy
		elif (hasattr(self.selection, 'portal')
				and self.selection.portal is not None):
			return self.selection.portal
		else:
			raise ValueError("Invalid selection: {}".format(self.selection))

	def refresh_selected_proxy(self, *args):
		self.selected_proxy = self._get_selected_proxy()

	def on_character_name(self, *args):
		if not self.engine:
			Clock.schedule_once(self.on_character_name, 0)
			return
		self.engine.eternal['boardchar'] = self.engine.character[
			self.character_name]

	def on_character(self, *args):
		if not hasattr(self, 'mainscreen'):
			Clock.schedule_once(self.on_character, 0)
			return
		if hasattr(self, '_oldchar'):
			self.mainscreen.graphboards[self._oldchar.name].unbind(
				selection=self.setter('selection'))
			self.mainscreen.gridboards[self._oldchar.name].unbind(
				selection=self.setter('selection'))
		self.selection = None
		self.mainscreen.graphboards[self.character.name].bind(
			selection=self.setter('selection'))
		self.mainscreen.gridboards[self.character.name].bind(
			selection=self.setter('selection'))

	def on_pause(self):
		"""Sync the database with the current state of the game."""
		if hasattr(self, 'engine'):
			self.engine.commit()
		self.strings.save()
		self.funcs.save()

	def on_stop(self, *largs):
		"""Sync the database, wrap up the game, and halt."""
		self.mainmenu.closed = True
		self.strings.save()
		self.funcs.save()
		if self.engine:
			self.engine.commit()
		if hasattr(self, 'procman'):
			self.procman.shutdown()
		if hasattr(self, 'starting_dir'):
			os.chdir(self.starting_dir)

	def delete_selection(self):
		"""Delete both the selected widget and whatever it represents."""
		selection = self.selection
		if selection is None:
			return
		if isinstance(selection, GraphArrowWidget):
			if selection.reciprocal and selection.reciprocal.portal.get(
				'is_mirror', False):
				selection.reciprocal.portal.delete()
				self.mainscreen.boardview.board.rm_arrow(
					selection.destination.name, selection.origin.name)
			self.mainscreen.boardview.board.rm_arrow(
				selection.origin.name, selection.destination.name)
			selection.portal.delete()
		elif isinstance(selection, GraphSpot):
			charn = selection.board.character.name
			self.mainscreen.graphboards[charn].rm_spot(selection.name)
			gridb = self.mainscreen.gridboards[charn]
			if selection.name in gridb.spot:
				gridb.rm_spot(selection.name)
			selection.proxy.delete()
		else:
			assert isinstance(selection, Pawn)
			charn = selection.board.character.name
			self.mainscreen.graphboards[charn].rm_pawn(selection.name)
			self.mainscreen.gridboards[charn].rm_pawn(selection.name)
			selection.proxy.delete()
		self.selection = None

	def new_board(self, name):
		"""Make a graph for a character name, and switch to it."""
		char = self.engine.character[name]
		self.mainscreen.graphboards[name] = GraphBoard(character=char)
		self.mainscreen.gridboards[name] = GridBoard(character=char)
		self.character = char

	def on_edit_locked(self, *args):
		Logger.debug("ELiDEApp: " + (
			"edit locked" if self.edit_locked else "edit unlocked"))



from collections import defaultdict
from threading import Thread

from ELiDE.util import trigger

from kivy.app import App
from kivy.clock import triggered
from kivy.lang import Builder
from kivy.logger import Logger
from kivy.properties import BooleanProperty, NumericProperty, ObjectProperty, \
 StringProperty
from kivy.graphics import Color, Line
from kivy.uix.recycleview import RecycleView
from kivy.uix.button import Button
from kivy.uix.screenmanager import Screen
from kivy.uix.widget import Widget


class ThornyRectangle(Button):
	left_margin = NumericProperty(10)
	right_margin = NumericProperty(10)
	top_margin = NumericProperty(10)
	bottom_margin = NumericProperty(10)

	draw_left = BooleanProperty(False)
	draw_right = BooleanProperty(False)
	draw_up = BooleanProperty(False)
	draw_down = BooleanProperty(False)

	branch = StringProperty()
	turn = NumericProperty()

	def __init__(self, **kwargs):
		super().__init__(**kwargs)
		self.bind(
			pos=self._trigger_redraw,
			size=self._trigger_redraw,
			left_margin=self._trigger_redraw,
			right_margin=self._trigger_redraw,
			top_margin=self._trigger_redraw,
			bottom_margin=self._trigger_redraw,
			draw_left=self._trigger_redraw,
			draw_right=self._trigger_redraw,
			draw_up=self._trigger_redraw,
			draw_down=self._trigger_redraw,
		)
		self._trigger_redraw()

	def collide_point(self, x, y):
		return (self.x + self.left_margin < x < self.right - self.right_margin
				and
				self.y + self.bottom_margin < y < self.top - self.top_margin)

	def _redraw_line(self, enabled, name, point_lambda):
		if enabled:
			points = point_lambda()
			if hasattr(self, name):
				the_line = getattr(self, name)
				the_line.points = points
			else:
				the_line = Line(points=points)
			if the_line not in self.canvas.children:
				self.canvas.add(the_line)
			setattr(self, name, the_line)
		elif hasattr(self, name):
			self.canvas.remove(getattr(self, name))
			delattr(self, name)

	def _get_left_line_points(self):
		return [
			self.x, self.center_y, self.x + self.left_margin, self.center_y
		]

	def _get_right_line_points(self):
		return [
			self.right - self.right_margin, self.center_y, self.right,
			self.center_y
		]

	def _get_top_line_points(self):
		return [
			self.center_x, self.top, self.center_x, self.top - self.top_margin
		]

	def _get_bottom_line_points(self):
		return [
			self.center_x, self.y, self.center_x, self.y + self.bottom_margin
		]

	def _redraw(self, *args):
		self._color = Color(rgba=[1, 1, 1, 1])
		if self._color not in self.canvas.children:
			self.canvas.add(self._color)
		rectpoints = [
			self.x + self.left_margin, self.y + self.bottom_margin,
			self.right - self.right_margin, self.y + self.bottom_margin,
			self.right - self.right_margin, self.top - self.top_margin,
			self.x + self.left_margin, self.top - self.top_margin,
			self.x + self.left_margin, self.y + self.bottom_margin
		]
		if hasattr(self, '_rect'):
			self._rect.points = rectpoints
		else:
			self._rect = Line(points=rectpoints)
			self.canvas.add(self._rect)
		self._redraw_line(self.draw_left, '_left_line',
							self._get_left_line_points)
		self._redraw_line(self.draw_right, '_right_line',
							self._get_right_line_points)
		self._redraw_line(self.draw_up, '_top_line', self._get_top_line_points)
		self._redraw_line(self.draw_down, '_bot_line',
							self._get_bottom_line_points)
		self.canvas.ask_update()

	_trigger_redraw = trigger(_redraw)

	def on_release(self):
		if self.branch is None or self.turn is None:
			return
		app = App.get_running_app()
		app.mainscreen.toggle_timestream()
		self._push_time()

	@triggered(timeout=0.1)
	def _push_time(self, *args):
		app = App.get_running_app()
		app.time_travel(self.branch, self.turn)


class Cross(Widget):
	draw_left = BooleanProperty(True)
	draw_right = BooleanProperty(True)
	draw_up = BooleanProperty(True)
	draw_down = BooleanProperty(True)

	def __init__(self, **kwargs):
		super().__init__(**kwargs)
		self.bind(draw_left=self._trigger_redraw,
					draw_right=self._trigger_redraw,
					draw_up=self._trigger_redraw,
					draw_down=self._trigger_redraw,
					size=self._trigger_redraw,
					pos=self._trigger_redraw)

	def _draw_line(self, enabled, name, get_points):
		if enabled:
			points = get_points()
			if hasattr(self, name):
				getattr(self, name).points = points
			else:
				the_line = Line(points=points)
				setattr(self, name, the_line)
				self.canvas.add(the_line)
		elif hasattr(self, name):
			the_line = getattr(self, name)
			if the_line in self.canvas.children:
				self.canvas.remove(the_line)
			delattr(self, name)

	def _get_left_points(self):
		return [self.x, self.center_y, self.center_x, self.center_y]

	def _get_right_points(self):
		return [self.center_x, self.center_y, self.right, self.center_y]

	def _get_up_points(self):
		return [self.center_x, self.center_y, self.center_x, self.top]

	def _get_down_points(self):
		return [self.center_x, self.center_y, self.center_x, self.y]

	def _redraw(self, *args):
		self._draw_line(self.draw_left, '_left_line', self._get_left_points)
		self._draw_line(self.draw_right, '_right_line', self._get_right_points)
		self._draw_line(self.draw_up, '_up_line', self._get_up_points)
		self._draw_line(self.draw_down, '_down_line', self._get_down_points)
		self.canvas.ask_update()

	_trigger_redraw = trigger(_redraw)


class Timestream(RecycleView):
	cols = NumericProperty(
		1
	)  # should be equal to the number of turns on which branches were created + 1


def _data_and_cols_from_branches(branches):
	start_turn_branches = defaultdict(set)
	end_turn_branches = defaultdict(set)
	branch_split_turns_todo = defaultdict(set)
	branch_split_turns_done = defaultdict(set)
	for branch, (parent, parent_turn, parent_tick, end_turn,
					end_tick) in branches.items():
		start_turn_branches[parent_turn].add(branch)
		end_turn_branches[end_turn].add(branch)
		branch_split_turns_todo[parent].add(parent_turn)
	branch_split_turns_todo['trunk'].add(0)
	col2turn = sorted(start_turn_branches.keys() | end_turn_branches.keys())
	if not col2turn:
		return [], 1
	Logger.debug("Timestream: read branch lineage, processing...")
	data = []
	trunk_lineage = branches.pop('trunk')
	sorted_branches = [('trunk', (None, 0, 0, 0, 0))] + sorted(
		branches.items(), key=lambda x: x[1][1])
	branches['trunk'] = trunk_lineage
	for row, (branch, _) in enumerate(sorted_branches):
		for turn in col2turn:
			if branch == 'trunk' and turn == 0:
				data.append({
					'widget': 'ThornyRectangle',
					'branch': 'trunk',
					'turn': 0,
					'draw_left': False,
					'draw_up': False,
					'draw_down': len(start_turn_branches[turn]) > 1,
					'draw_right': bool(branch_split_turns_todo[branch])
				})
			elif branch in start_turn_branches[turn]:
				here_branches = [(branches[b][1], b)
									for b in start_turn_branches[turn]]
				if branch == min(here_branches)[1]:
					data.append({
						'widget': 'ThornyRectangle',
						'branch': branch,
						'turn': turn,
						'draw_left': False,
						'draw_up': turn == branches[branch][1],
						'draw_down': len(start_turn_branches[turn]) > 1,
						'draw_right': branches[branch][3] > turn
					})
				elif branch == max(here_branches)[1]:
					data.append({
						'widget': 'ThornyRectangle',
						'branch': branch,
						'turn': turn,
						'draw_left': branches[branch][1] > turn,
						'draw_up': False,
						'draw_down': len(start_turn_branches[turn]) > 1,
						'draw_right': False
					})
				else:
					data.append({
						'widget': 'Cross',
						'draw_left': False,
						'draw_up': True,
						'draw_down': len(start_turn_branches[turn]) > 1,
						'draw_right': branches[branch][3] > turn
					})
			elif branch in end_turn_branches[turn]:
				data.append({
					'widget':
					'ThornyRectangle',
					'branch':
					branch,
					'turn':
					turn,
					'draw_left':
					True,
					'draw_up':
					row > 0 and bool(start_turn_branches[turn]),
					'draw_down':
					bool(start_turn_branches[turn]),
					'draw_right':
					False
				})
			elif branches[branch][1] <= turn:
				here_branches = [(branches[b][1], b)
									for b in start_turn_branches[turn]
									if b != branch]
				data.append({
					'widget':
					'Cross',
					'draw_left':
					turn < branches[branch][3],
					'draw_right':
					turn < branches[branch][3],
					'draw_up':
					branch in branches and branches[branch][0] in branches
					and branches[branches[branch][0]][3] >= turn,
					'draw_down':
					bool(here_branches)
				})
			else:
				data.append({'widget': 'Widget'})
			branch_split_turns_todo[branch].discard(turn)
			start_turn_branches[turn].discard(branch)
			if turn in end_turn_branches:
				end_turn_branches[turn].discard(branch)
			branch_split_turns_done[branch].add(turn)
		Logger.debug(f"Timestream: processed branch {branch}")
	return data, len(col2turn)


class TimestreamScreen(Screen):
	toggle = ObjectProperty()
	timestream = ObjectProperty()
	_thread: Thread

	def on_pre_enter(self, *args):
		self.timestream.disabled = True
		self._thread = Thread(target=self._get_data)
		self._thread.start()

	def _get_data(self, *args):
		Logger.debug("Timestream: getting branches")
		engine = App.get_running_app().engine
		data, cols = _data_and_cols_from_branches(engine.handle('branches'))
		self.timestream.cols = cols
		self.timestream.data = data
		Logger.debug("Timestream: loaded!")
		self.timestream.disabled = False


Builder.load_string(r"""
<ThornyRectangle>:
    text: f"{self.branch}\n{int(self.turn)}"
<Timestream>:
    key_viewclass: 'widget'
    effect_cls: 'ScrollEffect'
    RecycleGridLayout:
        cols: root.cols
        default_width: 100
        default_height: 100
        default_size_hint: None, None
        height: self.minimum_height
        width: self.minimum_width
        size_hint: None, None
<TimestreamScreen>:
    name: 'timestream'
    timestream: timestream
    BoxLayout:
        orientation: 'vertical'
        Timestream:
            id: timestream
            size_hint_y: 0.95
        BoxLayout:
            size_hint_y: 0.05
            Button:
                text: 'Cancel'
                on_press: root.toggle()
""")

if __name__ == "__main__":
	from kivy.base import runTouchApp
	branches = {
		'trunk': (None, 0, 0, 1, 27),
		'trunk1': ('trunk', 0, 2494, 0, 2494),
		'trunk2': ('trunk1', 0, 2494, 1, 27),
		'trunk3': ('trunk', 1, 27, 1, 31)
	}
	data, cols = _data_and_cols_from_branches(branches)
	timestream = Timestream(cols=cols)
	timestream.data = data
	runTouchApp(timestream)



from kivy.properties import (ObjectProperty, BooleanProperty, ListProperty,
								DictProperty)

from ..imagestackproxy import ImageStackProxy
from ..pawn import PawnBehavior


class GridPawn(ImageStackProxy, PawnBehavior):
	board = ObjectProperty()

	def __repr__(self):
		"""Give my ``thing``'s name and its location's name."""
		return '<{}-in-{} at {}>'.format(self.name, self.loc_name, id(self))
